__author__ = "sibirrer"

# this file contains a class to compute the Navaro-Frank-White function in mass/kappa space
# the potential therefore is its integral

import numpy as np
from lenstronomy.LensModel.Profiles.cnfw import CNFW
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CNFW_ELLIPSE"]


class CNFW_ELLIPSE(LensProfileBase):
    """This class contains functions concerning the NFW profile.

    relation are: R_200 = c * Rs
    """

    param_names = ["Rs", "alpha_Rs", "r_core", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "r_core": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "r_core": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.cnfw = CNFW()
        self._diff = 0.0000000001
        super(CNFW_ELLIPSE, self).__init__()

    def function(self, x, y, Rs, alpha_Rs, r_core, e1, e2, center_x=0, center_y=0):
        """Returns double integral of NFW profile."""
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = min(abs(1.0 - q), 0.99)
        xt1 = (cos_phi * x_shift + sin_phi * y_shift) * np.sqrt(1 - e)
        xt2 = (-sin_phi * x_shift + cos_phi * y_shift) * np.sqrt(1 + e)
        R_ = np.sqrt(xt1**2 + xt2**2)
        f_ = self.cnfw.function(R_, 0, Rs, alpha_Rs, r_core, center_x=0, center_y=0)
        return f_

    def derivatives(self, x, y, Rs, alpha_Rs, r_core, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (integral of NFW)"""
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = min(abs(1.0 - q), 0.99)
        xt1 = (cos_phi * x_shift + sin_phi * y_shift) * np.sqrt(1 - e)
        xt2 = (-sin_phi * x_shift + cos_phi * y_shift) * np.sqrt(1 + e)

        f_x_prim, f_y_prim = self.cnfw.derivatives(
            xt1, xt2, Rs, alpha_Rs, r_core, center_x=0, center_y=0
        )
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, Rs, alpha_Rs, r_core, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy."""
        diff = 0.0000001
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, Rs, alpha_Rs, r_core, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, Rs, alpha_Rs, r_core, e1, e2, center_x, center_y
        )

        alpha_ra_dx_, alpha_dec_dx_ = self.derivatives(
            x - diff, y, Rs, alpha_Rs, r_core, e1, e2, center_x, center_y
        )
        alpha_ra_dy_, alpha_dec_dy_ = self.derivatives(
            x, y - diff, Rs, alpha_Rs, r_core, e1, e2, center_x, center_y
        )

        dalpha_rara = (alpha_ra_dx - alpha_ra_dx_) / diff / 2
        dalpha_radec = (alpha_ra_dy - alpha_ra_dy_) / diff / 2
        dalpha_decra = (alpha_dec_dx - alpha_dec_dx_) / diff / 2
        dalpha_decdec = (alpha_dec_dy - alpha_dec_dy_) / diff / 2

        f_xx = dalpha_rara
        f_yy = dalpha_decdec
        f_xy = dalpha_radec
        f_yx = dalpha_decra
        return f_xx, f_xy, f_yx, f_yy

    def mass_3d_lens(self, R, Rs, alpha_Rs, r_core, e1=0, e2=0):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units.

        :return:
        """
        return self.cnfw.mass_3d_lens(R, Rs, alpha_Rs, r_core)

    def density_lens(self, R, Rs, alpha_Rs, r_core, e1=0, e2=0):
        """Computes the density at 3d radius r given lens model parameterization.

        The integral in the LOS projection of this quantity results in the convergence
        quantity.
        """
        return self.cnfw.density_lens(R, Rs, alpha_Rs, r_core)



import pytest
from lenstronomy.GalKin.analytic_kinematics import AnalyticKinematics


class TestAnalyticKinematics(object):
    def setup_method(self):
        pass

    def test_sigma_s2(self):
        kwargs_aperture = {
            "center_ra": 0,
            "width": 1,
            "length": 1,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": "slit",
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
        kin = AnalyticKinematics(kwargs_cosmo)
        kwargs_light = {"r_eff": 1}
        sigma_s2 = kin.sigma_s2(
            r=1,
            R=0.1,
            kwargs_mass={"theta_E": 1, "gamma": 2},
            kwargs_light=kwargs_light,
            kwargs_anisotropy={"r_ani": 1},
        )
        assert "a" in kwargs_light


if __name__ == "__main__":
    pytest.main()



import numpy as np
from lenstronomy.PointSource.Types.base_ps import PSBase, _expand_to_array

__all__ = ["LensedPositions"]


class LensedPositions(PSBase):
    """
    class of a lensed point source parameterized as the (multiple) observed image positions
    Name within the PointSource module: 'LENSED_POSITION'
    parameters: ra_image, dec_image, point_amp
    If fixed_magnification=True, than 'source_amp' is a parameter instead of 'point_amp'

    """

    # def __init__(self, lens_model=None, fixed_magnification=False, additional_image=False):
    #    super(LensedPositions, self).__init__(lens_model=lens_model, fixed_magnification=fixed_magnification,
    #                                          additional_image=additional_image)

    def image_position(
        self,
        kwargs_ps,
        kwargs_lens=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
        additional_images=False,
    ):
        """On-sky image positions.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            requiring the lens equation solver
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :param additional_images: if True, solves the lens equation for additional
            images
        :type additional_images: bool
        :return: image positions in x, y as arrays
        """
        if self.additional_images is True or additional_images:
            if kwargs_lens_eqn_solver is None:
                kwargs_lens_eqn_solver = {}
            ra_source, dec_source = self.source_position(kwargs_ps, kwargs_lens)
            # TODO: this solver does not distinguish between different frames/bands with partial lens models
            ra_image, dec_image = self._solver.image_position_from_source(
                ra_source,
                dec_source,
                kwargs_lens,
                magnification_limit=magnification_limit,
                **kwargs_lens_eqn_solver
            )
        else:
            ra_image = kwargs_ps["ra_image"]
            dec_image = kwargs_ps["dec_image"]
        return np.array(ra_image), np.array(dec_image)

    def source_position(self, kwargs_ps, kwargs_lens=None):
        """Original source position (prior to lensing)

        :param kwargs_ps: point source keyword arguments
        :param kwargs_lens: lens model keyword argument list (required to ray-trace back
            in the source plane)
        :return: x, y position (as numpy arrays)
        """
        ra_image = kwargs_ps["ra_image"]
        dec_image = kwargs_ps["dec_image"]

        if self.k_list is None:
            x_source, y_source = self._lens_model.ray_shooting(
                ra_image, dec_image, kwargs_lens
            )
        else:
            x_source, y_source = [], []
            for i in range(len(ra_image)):
                x, y = self._lens_model.ray_shooting(
                    ra_image[i], dec_image[i], kwargs_lens, k=self.k_list[i]
                )
                x_source.append(x)
                y_source.append(y)
        x_source = np.mean(x_source)
        y_source = np.mean(y_source)
        return np.array(x_source), np.array(y_source)

    def image_amplitude(
        self,
        kwargs_ps,
        kwargs_lens=None,
        x_pos=None,
        y_pos=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
    ):
        """Image brightness amplitudes.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            requiring the lens equation solver
        :param x_pos: pre-computed image position (no lens equation solver applied)
        :param y_pos: pre-computed image position (no lens equation solver applied)
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :return: array of image amplitudes
        """
        if self._fixed_magnification:
            if x_pos is not None and y_pos is not None:
                ra_image, dec_image = x_pos, y_pos
            else:
                ra_image, dec_image = self.image_position(
                    kwargs_ps,
                    kwargs_lens,
                    magnification_limit=magnification_limit,
                    kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
                )

            if self.k_list is None:
                mag = self._lens_model.magnification(ra_image, dec_image, kwargs_lens)
            else:
                mag = []
                for i in range(len(ra_image)):
                    mag.append(
                        self._lens_model.magnification(
                            ra_image, dec_image, kwargs_lens, k=self.k_list[i]
                        )
                    )
            point_amp = kwargs_ps["source_amp"] * np.abs(mag)
        else:
            point_amp = kwargs_ps["point_amp"]
            if x_pos is not None:
                point_amp = _expand_to_array(point_amp, len(x_pos))
        return np.array(point_amp)

    def source_amplitude(self, kwargs_ps, kwargs_lens=None):
        """Intrinsic brightness amplitude of point source When brightnesses are defined
        in magnified on-sky positions, the intrinsic brightness is computed as the mean
        in the magnification corrected image position brightnesses.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), used when
            brightness are defined in magnified on-sky positions
        :return: brightness amplitude (as numpy array)
        """
        if self._fixed_magnification:
            source_amp = kwargs_ps["source_amp"]
        else:
            ra_image, dec_image = kwargs_ps["ra_image"], kwargs_ps["dec_image"]
            if self.k_list is None:
                mag = self._lens_model.magnification(ra_image, dec_image, kwargs_lens)
            else:
                mag = []
                for i in range(len(ra_image)):
                    mag.append(
                        self._lens_model.magnification(
                            ra_image, dec_image, kwargs_lens, k=self.k_list[i]
                        )
                    )
            point_amp = kwargs_ps["point_amp"]
            source_amp = np.mean(np.array(point_amp) / np.array(np.abs(mag)))
        return np.array(source_amp)



__author__ = "sibirrer"


import lenstronomy.Util.correlation as correlation

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.Util import util


class TestCorrelation(object):
    def setup_method(self):
        pass

    def test_power_spectrum_2D(self):
        num_pix = 100
        x, y = util.make_grid(numPix=num_pix, deltapix=1)
        I_r = util.array2image(np.cos(x / num_pix / (2 * np.pi)))
        psd2D = correlation.power_spectrum_2d(I_r)
        npt.assert_almost_equal(np.max(psd2D), 1, decimal=1)

    def test_power_spectrum_1D(self):
        num_pix = 100
        np.random.seed(42)
        I = np.random.random((num_pix, num_pix))
        # I[5, 5] = 100
        # I[50, 5] = 100

        psd1D, r = correlation.power_spectrum_1d(I)
        # print(np.max(psd1D))
        # print(psd1D)

        # import matplotlib.pyplot as plt
        # plt.plot(psd1D)
        # plt.show()
        print(np.average(psd1D[: int(num_pix / 2.0)]))
        # this tests whether the white noise power-spectrum is flat:
        npt.assert_almost_equal(
            np.average(psd1D[: int(num_pix / 2.0)])
            / np.average(psd1D[int(num_pix / 2.0) :]),
            1,
            decimal=1,
        )

        num_pix = 10
        residuals = np.ones((num_pix, num_pix))
        residuals[5, 5] = num_pix**2
        psd1D, r = correlation.power_spectrum_1d(residuals)
        print(psd1D)
        npt.assert_almost_equal(
            psd1D, ((num_pix**2 - 1.0) / num_pix**2) ** 2, decimal=7
        )


if __name__ == "__main__":
    pytest.main()



import numpy as np
from lenstronomy.Cosmo.background import Background

__all__ = ["Image2SourceMapping"]


class Image2SourceMapping(object):
    """This class handles multiple source planes and performs the computation of
    predicted surface brightness at given image positions. The class is enable to deal
    with an arbitrary number of different source planes. There are two different
    settings:

    Single lens plane modelling:
    In case of a single deflector, lenstronomy models the reduced deflection angles
    (matched to the source plane in single source plane mode). Each source light model can be added a number
    (scale_factor) that rescales the reduced deflection angle to the specific source plane.

    Multiple lens plane modelling:
    The multi-plane lens modelling requires the assumption of a cosmology and the redshifts of the multiple lens and
    source planes. The backwards ray-tracing is performed and stopped at the different source plane redshift to compute
    the mapping between source to image plane.
    """

    def __init__(self, lensModel, sourceModel):
        """

        :param lensModel: LensModel() class instance
        :param sourceModel: LightModel() class instance.
         The lightModel includes:

         - source_scale_factor_list: list of floats corresponding to the rescaled deflection angles to the specific source components. None indicates that the list will be set to 1, meaning a single source plane model (in single lens plane mode).
         - source_redshift_list: list of redshifts of the light components (in multi lens plane mode)
        """

        self._lightModel = sourceModel
        self._lensModel = lensModel
        light_model_list = sourceModel.profile_type_list
        self._multi_lens_plane = lensModel.multi_plane
        self._source_redshift_list = sourceModel.redshift_list
        self._deflection_scaling_list = sourceModel.deflection_scaling_list
        self._multi_source_plane = True

        if self._multi_lens_plane is True:
            if self._deflection_scaling_list is not None:
                raise ValueError(
                    "deflection scaling for different source planes not possible in combination of "
                    "multi-lens plane modeling. You have to specify the redshifts of the sources instead."
                )
            self._bkg_cosmo = Background(lensModel.cosmo)
            if self._source_redshift_list is None:
                self._multi_source_plane = False
            elif len(self._source_redshift_list) != len(light_model_list):
                raise ValueError(
                    "length of redshift_list must correspond to length of light_model_list"
                )
            elif np.max(self._source_redshift_list) > self._lensModel.z_source:
                raise ValueError(
                    "redshift of source_redshift_list have to be smaler or equal to the one specified in "
                    "the lens model."
                )
            else:
                self._sorted_source_redshift_index = self._index_ordering(
                    self._source_redshift_list
                )
                self._T0z_list = []
                for z_stop in self._source_redshift_list:
                    self._T0z_list.append(self._bkg_cosmo.T_xy(0, z_stop))
                z_start = 0
                self._T_ij_start_list = []
                self._T_ij_end_list = []
                for i, index_source in enumerate(self._sorted_source_redshift_index):
                    z_stop = self._source_redshift_list[index_source]
                    (
                        T_ij_start,
                        T_ij_end,
                    ) = self._lensModel.lens_model.transverse_distance_start_stop(
                        z_start, z_stop, include_z_start=False
                    )
                    self._T_ij_start_list.append(T_ij_start)
                    self._T_ij_end_list.append(T_ij_end)
                    z_start = z_stop
        else:
            if self._deflection_scaling_list is None:
                self._multi_source_plane = False
            elif len(self._deflection_scaling_list) != len(light_model_list):
                raise ValueError(
                    "length of scale_factor_list must correspond to length of light_model_list!"
                )

    def image2source(self, x, y, kwargs_lens, index_source):
        """
        mapping of image plane to source plane coordinates
        WARNING: for multi lens plane computations and multi source planes, this computation can be slow and should be
        used as rarely as possible.

        :param x: image plane coordinate (angle)
        :param y: image plane coordinate (angle)
        :param kwargs_lens: lens model kwargs list
        :param index_source: int, index of source model
        :return: source plane coordinate corresponding to the source model of index idex_source
        """
        if self._multi_source_plane is False:
            x_source, y_source = self._lensModel.ray_shooting(x, y, kwargs_lens)
        else:
            if self._multi_lens_plane is False:
                x_alpha, y_alpha = self._lensModel.alpha(x, y, kwargs_lens)
                scale_factor = self._deflection_scaling_list[index_source]
                x_source = x - x_alpha * scale_factor
                y_source = y - y_alpha * scale_factor
            else:
                z_stop = self._source_redshift_list[index_source]
                x_ = np.zeros_like(x)
                y_ = np.zeros_like(y)
                (
                    x_comov,
                    y_comov,
                    alpha_x,
                    alpha_y,
                ) = self._lensModel.lens_model.ray_shooting_partial(
                    x_, y_, x, y, 0, z_stop, kwargs_lens, include_z_start=False
                )

                T_z = self._T0z_list[index_source]
                x_source = x_comov / T_z
                y_source = y_comov / T_z
        return x_source, y_source

    def image_flux_joint(self, x, y, kwargs_lens, kwargs_source, k=None):
        """

        :param x: coordinate in image plane
        :param y: coordinate in image plane
        :param kwargs_lens: lens model kwargs list
        :param kwargs_source: source model kwargs list
        :param k: None or int or list of int for partial evaluation of light models
        :return: surface brightness of all joint light components at image position (x, y)
        """
        if self._multi_source_plane is False:
            x_source, y_source = self._lensModel.ray_shooting(x, y, kwargs_lens)
            return self._lightModel.surface_brightness(
                x_source, y_source, kwargs_source, k=k
            )
        else:
            flux = np.zeros_like(x)
            if self._multi_lens_plane is False:
                x_alpha, y_alpha = self._lensModel.alpha(x, y, kwargs_lens)
                for i in range(len(self._deflection_scaling_list)):
                    scale_factor = self._deflection_scaling_list[i]
                    x_source = x - x_alpha * scale_factor
                    y_source = y - y_alpha * scale_factor
                    if k is None or k == i:
                        flux += self._lightModel.surface_brightness(
                            x_source, y_source, kwargs_source, k=i
                        )
            else:
                x_comov = np.zeros_like(x)
                y_comov = np.zeros_like(y)
                alpha_x, alpha_y = x, y
                x_source, y_source = alpha_x, alpha_y
                z_start = 0
                for i, index_source in enumerate(self._sorted_source_redshift_index):
                    z_stop = self._source_redshift_list[index_source]
                    if z_stop > z_start:
                        T_ij_start = self._T_ij_start_list[i]
                        T_ij_end = self._T_ij_end_list[i]
                        (
                            x_comov,
                            y_comov,
                            alpha_x,
                            alpha_y,
                        ) = self._lensModel.lens_model.ray_shooting_partial(
                            x_comov,
                            y_comov,
                            alpha_x,
                            alpha_y,
                            z_start,
                            z_stop,
                            kwargs_lens,
                            include_z_start=False,
                            T_ij_start=T_ij_start,
                            T_ij_end=T_ij_end,
                        )

                        T_z = self._T0z_list[index_source]
                        x_source = x_comov / T_z
                        y_source = y_comov / T_z
                    if k is None or k == i:
                        flux += self._lightModel.surface_brightness(
                            x_source, y_source, kwargs_source, k=index_source
                        )
                    z_start = z_stop
            return flux

    def image_flux_split(self, x, y, kwargs_lens, kwargs_source):
        """

        :param x: coordinate in image plane
        :param y: coordinate in image plane
        :param kwargs_lens: lens model kwargs list
        :param kwargs_source: source model kwargs list
        :return: list of responses of every single basis component with default amplitude amp=1, in the same order as the light_model_list
        """
        if self._multi_source_plane is False:
            x_source, y_source = self._lensModel.ray_shooting(x, y, kwargs_lens)
            return self._lightModel.functions_split(x_source, y_source, kwargs_source)
        else:
            response = []
            n = 0
            if self._multi_lens_plane is False:
                x_alpha, y_alpha = self._lensModel.alpha(x, y, kwargs_lens)
                for i in range(len(self._deflection_scaling_list)):
                    scale_factor = self._deflection_scaling_list[i]
                    x_source = x - x_alpha * scale_factor
                    y_source = y - y_alpha * scale_factor
                    response_i, n_i = self._lightModel.functions_split(
                        x_source, y_source, kwargs_source, k=i
                    )
                    response += response_i
                    n += n_i
            else:
                n_i_list = []
                x_comov = np.zeros_like(x)
                y_comov = np.zeros_like(y)
                alpha_x, alpha_y = x, y
                x_source, y_source = alpha_x, alpha_y
                z_start = 0
                for i, index_source in enumerate(self._sorted_source_redshift_index):
                    z_stop = self._source_redshift_list[index_source]
                    if z_stop > z_start:
                        T_ij_start = self._T_ij_start_list[i]
                        T_ij_end = self._T_ij_end_list[i]
                        (
                            x_comov,
                            y_comov,
                            alpha_x,
                            alpha_y,
                        ) = self._lensModel.lens_model.ray_shooting_partial(
                            x_comov,
                            y_comov,
                            alpha_x,
                            alpha_y,
                            z_start,
                            z_stop,
                            kwargs_lens,
                            include_z_start=False,
                            T_ij_start=T_ij_start,
                            T_ij_end=T_ij_end,
                        )
                        T_z = self._T0z_list[index_source]
                        x_source = x_comov / T_z
                        y_source = y_comov / T_z
                    response_i, n_i = self._lightModel.functions_split(
                        x_source, y_source, kwargs_source, k=index_source
                    )
                    n_i_list.append(n_i)
                    response += response_i
                    n += n_i
                    z_start = z_stop
                n_list = self._lightModel.num_param_linear_list(kwargs_source)
                response = self._re_order_split(response, n_list)

            return response, n

    @staticmethod
    def _index_ordering(redshift_list):
        """

        :param redshift_list: list of redshifts
        :return: indexes in ascending order to be evaluated (from z=0 to z=z_source)
        """
        redshift_list = np.array(redshift_list)
        sort_index = np.argsort(redshift_list)
        return sort_index

    def _re_order_split(self, response, n_list):
        """

        :param response: splitted functions in order of redshifts
        :param n_list: list of number of response vectors per model in order of the model list (not redshift ordered)
        :return: reshuffled array in order of the function definition
        """
        counter_regular = 0
        n_sum_list_regular = []

        for i in range(len(self._source_redshift_list)):
            n_sum_list_regular += [counter_regular]
            counter_regular += n_list[i]

        reshuffled = np.zeros_like(response)
        n_sum_sorted = 0
        for i, index in enumerate(self._sorted_source_redshift_index):
            n_i = n_list[index]
            n_sum = n_sum_list_regular[index]
            reshuffled[n_sum : n_sum + n_i] = response[
                n_sum_sorted : n_sum_sorted + n_i
            ]
            n_sum_sorted += n_i
        return reshuffled



from lenstronomy.Sampling.Likelihoods.time_delay_likelihood import TimeDelayLikelihood
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from astropy.cosmology import FlatLambdaCDM
import numpy.testing as npt
import numpy as np
import pytest
import copy


class TestImageLikelihood(object):
    def setup_method(self):
        pass

    def test_logL(self):
        z_source = 1.5
        z_lens = 0.5
        cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.0)
        lensCosmo = LensCosmo(cosmo=cosmo, z_lens=z_lens, z_source=z_source)

        # make class instances for a chosen lens model type

        # chose a lens model
        lens_model_list = ["SPEP", "SHEAR"]
        # make instance of LensModel class
        lensModel = LensModel(
            lens_model_list=lens_model_list,
            cosmo=cosmo,
            z_lens=z_lens,
            z_source=z_source,
        )
        # we require routines accessible in the LensModelExtensions class
        # make instance of LensEquationSolver to solve the lens equation
        lensEquationSolver = LensEquationSolver(lensModel=lensModel)

        # make choice of lens model

        # we chose a source position (in units angle)
        x_source, y_source = 0.02, 0.01
        # we chose a lens model
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "e1": 0.1,
                "e2": 0.2,
                "gamma": 2.0,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.05, "gamma2": -0.01},
        ]

        # compute image positions and their (finite) magnifications

        # we solve for the image position(s) of the provided source position and lens model
        x_img, y_img = lensEquationSolver.image_position_from_source(
            kwargs_lens=kwargs_lens, sourcePos_x=x_source, sourcePos_y=y_source
        )

        point_source_list = ["LENSED_POSITION"]
        kwargs_ps = [{"ra_image": x_img, "dec_image": y_img}]
        pointSource = PointSource(point_source_type_list=point_source_list)
        t_days = lensModel.arrival_time(x_img, y_img, kwargs_lens)
        time_delays_measured = t_days[1:] - t_days[0]
        time_delays_uncertainties = np.array([0.1, 0.1, 0.1])
        self.td_likelihood = TimeDelayLikelihood(
            time_delays_measured,
            time_delays_uncertainties,
            lens_model_class=lensModel,
            point_source_class=pointSource,
        )
        kwargs_cosmo = {"D_dt": lensCosmo.ddt}
        logL = self.td_likelihood.logL(
            kwargs_lens=kwargs_lens, kwargs_ps=kwargs_ps, kwargs_cosmo=kwargs_cosmo
        )
        npt.assert_almost_equal(logL, 0, decimal=8)

        time_delays_measured_new = copy.deepcopy(time_delays_measured)
        time_delays_measured_new[0] += 0.1
        td_likelihood = TimeDelayLikelihood(
            time_delays_measured_new,
            time_delays_uncertainties,
            lens_model_class=lensModel,
            point_source_class=pointSource,
        )
        kwargs_cosmo = {"D_dt": lensCosmo.ddt}
        logL = td_likelihood.logL(
            kwargs_lens=kwargs_lens, kwargs_ps=kwargs_ps, kwargs_cosmo=kwargs_cosmo
        )
        npt.assert_almost_equal(logL, -0.5, decimal=8)

        # Test a covariance matrix being used
        time_delays_cov = np.diag([0.1, 0.1, 0.1]) ** 2
        td_likelihood = TimeDelayLikelihood(
            time_delays_measured_new,
            time_delays_cov,
            lens_model_class=lensModel,
            point_source_class=pointSource,
        )
        logL = td_likelihood.logL(
            kwargs_lens=kwargs_lens, kwargs_ps=kwargs_ps, kwargs_cosmo=kwargs_cosmo
        )
        npt.assert_almost_equal(logL, -0.5, decimal=8)

        # Test behaviour with a wrong number of images
        time_delays_measured_new = time_delays_measured_new[:-1]
        time_delays_uncertainties = time_delays_uncertainties[:-1]  # remove last image
        td_likelihood = TimeDelayLikelihood(
            time_delays_measured_new,
            time_delays_uncertainties,
            lens_model_class=lensModel,
            point_source_class=pointSource,
        )
        logL = td_likelihood.logL(
            kwargs_lens=kwargs_lens, kwargs_ps=kwargs_ps, kwargs_cosmo=kwargs_cosmo
        )
        npt.assert_almost_equal(logL, -(10**15), decimal=8)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from lenstronomy.Util import constants

# routines to conveniently compute micro-lensing events


def einstein_radius(mass, d_l, d_s):
    """Einstein radius for a given point mass and distances to lens and source.

    :param mass: point source mass [M_sun]
    :param d_l: distance to lens [pc]
    :param d_s: distance to source [pc]
    :return: Einstein radius [arc seconds]
    """
    mass_kg = mass * constants.M_sun
    dl_m = d_l * constants.pc
    ds_m = d_s * constants.pc
    # Einstein radius in radian
    theta_e = np.sqrt(
        4 * constants.G * mass_kg / constants.c**2 * (ds_m - dl_m) / (ds_m * dl_m)
    )
    theta_e /= constants.arcsec  # arc seconds
    return theta_e


def source_size(diameter, d_s):
    """

    :param diameter: diameter of the source in units of the solar diameter
    :param d_s: distance to the source in [pc]
    :return: diameter in [arc seconds]
    """
    diameter_m = diameter * constants.r_sun * 2  # diameter in [m]
    diameter_arcsec = diameter_m / (d_s * constants.pc)  # diameter in [radian]
    diameter_arcsec /= constants.arcsec  # diameter in [arc seconds]
    return diameter_arcsec



__author__ = "sibirrer"

import pytest
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.Plots import lens_plot
import lenstronomy.Plots.model_plot as output_plots

import matplotlib

matplotlib.use("agg")
import matplotlib.pyplot as plt


class TestLensPlot(object):
    """Test the fitting sequences."""

    def setup_method(self):
        pass

    def test_lens_model_plot(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["SIS"])
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        lens_plot.lens_model_plot(
            ax,
            lensModel,
            kwargs_lens,
            numPix=10,
            deltaPix=0.5,
            sourcePos_x=0,
            sourcePos_y=0,
            point_source=True,
            with_caustics=True,
            fast_caustic=False,
        )
        plt.close()

        lens_plot.lens_model_plot(
            ax,
            lensModel,
            kwargs_lens,
            numPix=10,
            deltaPix=0.5,
            sourcePos_x=0,
            sourcePos_y=0,
            point_source=True,
            with_caustics=True,
            fast_caustic=True,
        )
        plt.close()

        lens_plot.lens_model_plot(
            ax,
            lensModel,
            kwargs_lens,
            numPix=10,
            deltaPix=0.5,
            sourcePos_x=0,
            sourcePos_y=0,
            point_source=True,
            with_caustics=True,
            fast_caustic=True,
            coord_inverse=True,
        )
        plt.close()

    def test_arrival_time_surface(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["SIS"])
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        lens_plot.arrival_time_surface(
            ax,
            lensModel,
            kwargs_lens,
            numPix=100,
            deltaPix=0.05,
            sourcePos_x=0.02,
            sourcePos_y=0,
            point_source=True,
            with_caustics=True,
            image_color_list=["k", "k", "k", "r"],
        )
        plt.close()
        lens_plot.arrival_time_surface(
            ax,
            lensModel,
            kwargs_lens,
            numPix=100,
            deltaPix=0.05,
            sourcePos_x=0.02,
            sourcePos_y=0,
            point_source=True,
            with_caustics=False,
            image_color_list=None,
        )
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["SIS"])
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        lens_plot.arrival_time_surface(
            ax,
            lensModel,
            kwargs_lens,
            numPix=100,
            deltaPix=0.05,
            sourcePos_x=0.02,
            sourcePos_y=0,
            point_source=False,
            with_caustics=False,
        )
        plt.close()

    def test_distortions(self):
        lensModel = LensModel(lens_model_list=["SIS"])
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        lens_plot.distortions(
            lensModel,
            kwargs_lens,
            num_pix=10,
            delta_pix=0.2,
            center_ra=0,
            center_dec=0,
            differential_scale=0.0001,
        )
        plt.close()

        lens_plot.distortions(
            lensModel,
            kwargs_lens,
            num_pix=10,
            delta_pix=0.2,
            center_ra=0,
            center_dec=0,
            differential_scale=0.0001,
            smoothing_scale=0.1,
        )
        plt.close()

    def test_curved_arc_illustration(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["CURVED_ARC_SIS_MST"])
        kwargs_lens = [
            {
                "radial_stretch": 1.0466690706465702,
                "tangential_stretch": 4.598552192305616,
                "curvature": 0.8116297351731543,
                "direction": 2.6288852083221323,
                "center_x": -1.200866007937402,
                "center_y": 0.6829881436542166,
            }
        ]
        lens_plot.curved_arc_illustration(ax, lensModel, kwargs_lens)
        plt.close()

    def test_stretch_plot(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["SIE"])
        kwargs_lens = [
            {"theta_E": 1, "e1": 0.2, "e2": 0.0, "center_x": 0, "center_y": 0}
        ]
        lens_plot.stretch_plot(ax, lensModel, kwargs_lens)
        plt.close()

    def test_shear_plot(self):
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensModel = LensModel(lens_model_list=["SIE"])
        kwargs_lens = [
            {"theta_E": 1, "e1": 0.2, "e2": 0.0, "center_x": 0, "center_y": 0}
        ]
        lens_plot.shear_plot(ax, lensModel, kwargs_lens)
        plt.close()


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Hessian"]


class Hessian(LensProfileBase):
    """Class for constant Hessian distortion (second order) The input is in the same
    convention as the LensModel.hessian() output."""

    param_names = ["f_xx", "f_yy", "f_xy", "f_yx", "ra_0", "dec_0"]
    lower_limit_default = {
        "f_xx": -100,
        "f_yy": -100,
        "f_xy": -100,
        "f_yx": -100,
        "ra_0": -100,
        "dec_0": -100,
    }
    upper_limit_default = {
        "f_xx": 100,
        "f_yy": 100,
        "f_xy": 100,
        "f_yx": 100,
        "ra_0": 100,
        "dec_0": 100,
    }

    def function(self, x, y, f_xx, f_yy, f_xy, f_yx, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param f_xx: dalpha_x/dx
        :param f_yy: dalpha_y/dy
        :param f_xy: dalpha_x/dy
        :param f_yx: dalpha_y/dx
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: lensing potential
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_ = 1 / 2.0 * (f_xx * x_ * x_ + (f_xy + f_yx) * x_ * y_ + f_yy * y_ * y_)
        return f_

    def derivatives(self, x, y, f_xx, f_yy, f_xy, f_yx, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param f_xx: dalpha_x/dx
        :param f_yy: dalpha_y/dy
        :param f_xy: dalpha_x/dy
        :param f_yx: dalpha_y/dx
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: deflection angles
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_x = f_xx * x_ + f_xy * y_
        f_y = f_yx * x_ + f_yy * y_
        return f_x, f_y

    def hessian(self, x, y, f_xx, f_yy, f_xy, f_yx, ra_0=0, dec_0=0):
        """Hessian. Attention: If f_xy != f_yx then this function is not accurate!

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param f_xx: dalpha_x/dx
        :param f_yy: dalpha_y/dy
        :param f_xy: dalpha_x/dy
        :param f_yx: dalpha_y/dx
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: f_xx, f_yy, f_xy
        """
        return f_xx, f_xy, f_yx, f_yy



import numpy as np
import numpy.testing as npt
import pytest
import unittest

from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.Analysis.lens_profile import LensProfileAnalysis
from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import MultiGaussianKappa
import lenstronomy.Util.param_util as param_util


class TestLensProfileAnalysis(object):
    def setup_method(self):
        pass

    def test_profile_slope(self):
        lens_model = LensProfileAnalysis(LensModel(lens_model_list=["SPP"]))
        gamma_in = 2.0
        kwargs_lens = [
            {"theta_E": 1.0, "gamma": gamma_in, "center_x": 0, "center_y": 0}
        ]
        gamma_out = lens_model.profile_slope(kwargs_lens, radius=1)
        npt.assert_array_almost_equal(gamma_out, gamma_in, decimal=3)
        gamma_in = 1.7
        kwargs_lens = [
            {"theta_E": 1.0, "gamma": gamma_in, "center_x": 0, "center_y": 0}
        ]
        gamma_out = lens_model.profile_slope(kwargs_lens, radius=1)
        npt.assert_array_almost_equal(gamma_out, gamma_in, decimal=3)

        gamma_in = 2.5
        kwargs_lens = [
            {"theta_E": 1.0, "gamma": gamma_in, "center_x": 0, "center_y": 0}
        ]
        gamma_out = lens_model.profile_slope(kwargs_lens, radius=1)
        npt.assert_array_almost_equal(gamma_out, gamma_in, decimal=3)

        lens_model = LensProfileAnalysis(LensModel(lens_model_list=["SPEP"]))
        gamma_in = 2.0
        phi, q = 0.34403343049704888, 0.89760957136967312
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens = [
            {
                "theta_E": 1.4516812130749424,
                "e1": e1,
                "e2": e2,
                "center_x": -0.04507598845306314,
                "center_y": 0.054491803177414651,
                "gamma": gamma_in,
            }
        ]
        gamma_out = lens_model.profile_slope(kwargs_lens, radius=1.45)
        npt.assert_array_almost_equal(gamma_out, gamma_in, decimal=3)

    def test_effective_einstein_radius(self):
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        lensModel = LensProfileAnalysis(LensModel(lens_model_list=["SIS"]))
        ret = lensModel.effective_einstein_radius_grid(kwargs_lens, get_precision=True)

        assert len(ret) == 2
        npt.assert_almost_equal(ret[0], 1.0, decimal=2)
        theta_E = lensModel.effective_einstein_radius(
            kwargs_lens, r_min=1e-3, r_max=1e1, num_points=30
        )
        npt.assert_almost_equal(theta_E, 1, decimal=3)

        kwargs_lens_bad = [{"theta_E": 100, "center_x": 0, "center_y": 0}]
        ret_nan, precision = lensModel.effective_einstein_radius_grid(
            kwargs_lens_bad, get_precision=True
        )
        assert np.isnan(ret_nan)
        theta_E = lensModel.effective_einstein_radius(
            kwargs_lens_bad, r_min=1e-3, r_max=1e1, num_points=30
        )
        assert np.isnan(theta_E)

        # test interpolated profile
        numPix = 101
        deltaPix = 0.02
        from lenstronomy.Util import util

        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        from lenstronomy.LensModel.Profiles.sis import SIS

        sis = SIS()
        center_x, center_y = 0.0, -0.0
        kwargs_SIS = {"theta_E": 1.0, "center_x": center_x, "center_y": center_y}
        f_ = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x, f_y = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx, f_xy, f_yx, f_yy = sis.hessian(x_grid_interp, y_grid_interp, **kwargs_SIS)
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        kwargs_interpol = [
            {
                "grid_interp_x": x_axes,
                "grid_interp_y": y_axes,
                "f_": util.array2image(f_),
                "f_x": util.array2image(f_x),
                "f_y": util.array2image(f_y),
                "f_xx": util.array2image(f_xx),
                "f_xy": util.array2image(f_xy),
                "f_yy": util.array2image(f_yy),
            }
        ]
        lensModel = LensProfileAnalysis(LensModel(lens_model_list=["INTERPOL"]))
        theta_E_return = lensModel.effective_einstein_radius_grid(
            kwargs_interpol,
            get_precision=False,
            verbose=True,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(theta_E_return, 1, decimal=2)

        # sub-critical mass profile
        lensModel = LensProfileAnalysis(LensModel(lens_model_list=["NFW"]))
        kwargs_nfw = [{"Rs": 1, "alpha_Rs": 0.2, "center_x": 0, "center_y": 0}]
        theta_E_subcrit = lensModel.effective_einstein_radius_grid(
            kwargs_nfw, get_precision=False
        )
        assert np.isnan(theta_E_subcrit)
        theta_E_subcrit, _ = lensModel.effective_einstein_radius_grid(
            kwargs_nfw, get_precision=True
        )
        assert np.isnan(theta_E_subcrit)

    def test_external_lensing_effect(self):
        lens_model_list = ["SHEAR"]
        kwargs_lens = [{"gamma1": 0.1, "gamma2": 0.01}]
        lensModel = LensProfileAnalysis(LensModel(lens_model_list))
        alpha0_x, alpha0_y, kappa_ext, shear1, shear2 = lensModel.local_lensing_effect(
            kwargs_lens, model_list_bool=[0]
        )
        print(alpha0_x, alpha0_y, kappa_ext, shear1, shear2)
        assert alpha0_x == 0
        assert alpha0_y == 0
        assert shear1 == 0.1
        assert shear2 == 0.01
        assert kappa_ext == 0

    def test_multi_gaussian_lens(self):
        kwargs_options = {"lens_model_list": ["SPEP"]}
        lensModel = LensModel(**kwargs_options)
        lensAnalysis = LensProfileAnalysis(lens_model=lensModel)
        e1, e2 = param_util.phi_q2_ellipticity(0, 0.9)
        kwargs_lens = [
            {
                "gamma": 1.8,
                "theta_E": 0.6,
                "e1": e1,
                "e2": e2,
                "center_x": 0.5,
                "center_y": -0.1,
            }
        ]
        amplitudes, sigmas, center_x, center_y = lensAnalysis.multi_gaussian_lens(
            kwargs_lens, n_comp=20
        )
        model = MultiGaussianKappa()
        x = np.logspace(-2, 0.5, 10) + 0.5
        y = np.zeros_like(x) - 0.1
        f_xx, fxy, fyx, f_yy = model.hessian(
            x, y, amplitudes, sigmas, center_x=0.5, center_y=-0.1
        )
        kappa_mge = (f_xx + f_yy) / 2
        kappa_true = lensAnalysis._lens_model.kappa(x, y, kwargs_lens)
        print(kappa_true / kappa_mge)
        for i in range(len(x)):
            npt.assert_almost_equal(kappa_mge[i] / kappa_true[i], 1, decimal=1)

    def test_mass_fraction_within_radius(self):
        center_x, center_y = 0.5, -1
        theta_E = 1.1
        kwargs_lens = [{"theta_E": 1.1, "center_x": center_x, "center_y": center_y}]
        lensModel = LensModel(**{"lens_model_list": ["SIS"]})
        lensAnalysis = LensProfileAnalysis(lens_model=lensModel)
        kappa_mean_list = lensAnalysis.mass_fraction_within_radius(
            kwargs_lens, center_x, center_y, theta_E, numPix=100
        )
        npt.assert_almost_equal(kappa_mean_list[0], 1, 2)

    def test_lens_center(self):
        center_x, center_y = 0.43, -0.67
        kwargs_lens = [{"theta_E": 1, "center_x": center_x, "center_y": center_y}]
        lensModel = LensModel(**{"lens_model_list": ["SIS"]})
        profileAnalysis = LensProfileAnalysis(lens_model=lensModel)
        center_x_out, center_y_out = profileAnalysis.convergence_peak(kwargs_lens)
        npt.assert_almost_equal(center_x_out, center_x, 2)
        npt.assert_almost_equal(center_y_out, center_y, 2)

    def test_mst_invariant_differential(self):
        # testing with a SPP profile
        lensModel = LensModel(**{"lens_model_list": ["SPP"]})
        profileAnalysis = LensProfileAnalysis(lens_model=lensModel)
        gamma_list = [2, 1.8, 2.2]
        theta_E_list = [1, 0.5, 2]
        for gamma in gamma_list:
            for theta_E in theta_E_list:
                kwargs_lens = [
                    {"theta_E": theta_E, "gamma": gamma, "center_x": 0, "center_y": 0}
                ]
                xi = profileAnalysis.mst_invariant_differential(
                    kwargs_lens,
                    radius=theta_E,
                    center_x=None,
                    center_y=None,
                    model_list_bool=None,
                    num_points=10,
                )

                xi_true = (gamma - 2) / theta_E
                npt.assert_almost_equal(xi, xi_true, decimal=3)

        # it should also work in the elliptical regime
        lensModel = LensModel(**{"lens_model_list": ["EPL"]})
        profileAnalysis = LensProfileAnalysis(lens_model=lensModel)
        gamma_list = [2, 1.8, 2.2]
        theta_E_list = [1, 0.5, 2]
        q_list = [1, 0.9, 0.8, 0.7]
        for gamma in gamma_list:
            for theta_E in theta_E_list:
                for q in q_list:
                    e1, e2 = param_util.phi_q2_ellipticity(phi=0, q=q)
                    kwargs_lens = [
                        {
                            "theta_E": theta_E,
                            "gamma": gamma,
                            "e1": e1,
                            "e2": e2,
                            "center_x": 0,
                            "center_y": 0,
                        }
                    ]

                    theta_E_prime = np.sqrt(2 * q / (1 + q**2)) * theta_E
                    xi = profileAnalysis.mst_invariant_differential(
                        kwargs_lens,
                        radius=theta_E_prime,
                        center_x=None,
                        center_y=None,
                        model_list_bool=None,
                        num_points=10,
                    )
                    print(theta_E, gamma, q, "test")
                    xi_true = (gamma - 2) / theta_E_prime
                    npt.assert_almost_equal(xi, xi_true, decimal=2)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            raise ValueError()


if __name__ == "__main__":
    pytest.main()



import unittest
from lenstronomy.SimulationAPI.ObservationConfig.HST import HST
from lenstronomy.SimulationAPI.observation_api import Instrument, SingleBand
import lenstronomy.Util.util as util


class TestHST(unittest.TestCase):
    def setUp(self):
        self.TDLMC_F160W = HST()  # default is TDLMC_F160W
        self.F160W = HST(band="F160W")
        self.F160W2 = HST(band="F160W", psf_type="GAUSSIAN")

        kwargs_TDLMC_F160W = self.TDLMC_F160W.kwargs_single_band()
        kwargs_F160W = self.F160W.kwargs_single_band()
        kwargs_F160W2 = self.F160W2.kwargs_single_band()

        self.TDLMC_F160W_band = SingleBand(**kwargs_TDLMC_F160W)
        self.F160W_band = SingleBand(**kwargs_F160W)
        self.F160W2_band = SingleBand(**kwargs_F160W2)

        # dictionaries mapping HST kwargs to SingleBand kwargs
        self.camera_settings = {
            "read_noise": "_read_noise",
            "pixel_scale": "pixel_scale",
            "ccd_gain": "ccd_gain",
        }
        self.obs_settings = {
            "exposure_time": "_exposure_time",
            "sky_brightness": "_sky_brightness_",
            "magnitude_zero_point": "_magnitude_zero_point",
            "num_exposures": "_num_exposures",
            "seeing": "_seeing",
            "psf_type": "_psf_type",
        }

        self.instrument = Instrument(**self.TDLMC_F160W.camera)

    def test_HST_class(self):  # TODO: update; also text pixel/gaussian
        default = self.TDLMC_F160W
        explicit_TDLMC_F160W = HST(band="TDLMC_F160W", psf_type="PIXEL")
        self.assertEqual(explicit_TDLMC_F160W.camera, default.camera)
        self.assertEqual(explicit_TDLMC_F160W.obs, default.obs)

        with self.assertRaises(ValueError):
            bad_band = HST(band="g")

        with self.assertRaises(ValueError):
            bad_psf = HST(psf_type="blah")

        with self.assertRaises(ValueError):
            bad_coadd_years = HST(coadd_years=100)

    def test_HST_camera(self):
        # comparing camera settings in HST instance with those in Instrument instance
        for config, setting in self.camera_settings.items():
            self.assertEqual(
                self.TDLMC_F160W.camera[config],
                getattr(self.instrument, setting),
                msg=f"{config} did not match",
            )

    def test_HST_obs(self):
        # comparing obs settings in HST instance with those in SingleBand instance
        for config, setting in self.obs_settings.items():
            self.assertEqual(
                self.TDLMC_F160W.obs[config],
                getattr(self.TDLMC_F160W_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F160W.obs[config],
                getattr(self.F160W_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.F160W2.obs[config],
                getattr(self.F160W2_band, setting),
                msg=f"{config} did not match",
            )

    def test_kwargs_single_band(self):
        kwargs_TDLMC_F160W = util.merge_dicts(
            self.TDLMC_F160W.camera, self.TDLMC_F160W.obs
        )
        self.assertEqual(self.TDLMC_F160W.kwargs_single_band(), kwargs_TDLMC_F160W)


if __name__ == "__main__":
    unittest.main()



"""Tests for `galkin` module."""
import pytest
import numpy.testing as npt
import numpy as np
from lenstronomy.GalKin.light_profile import LightProfile
from lenstronomy.Analysis.light_profile import LightProfileAnalysis
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.GalKin import velocity_util
import scipy.integrate as integrate


class TestLightProfile(object):
    def setup_method(self):
        pass

    def test_draw_light(self):
        np.random.seed(41)
        lightProfile = LightProfile(profile_list=["HERNQUIST"])
        kwargs_profile = [{"amp": 1.0, "Rs": 0.8}]
        r_list = lightProfile.draw_light_2d(kwargs_profile, n=500000)
        bins = np.linspace(0.0, 1, 20)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        light2d = lightProfile.light_2d(
            R=(bins_hist[1:] + bins_hist[:-1]) / 2.0, kwargs_list=kwargs_profile
        )
        light2d *= (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light2d /= np.sum(light2d)
        hist /= np.sum(hist)
        print(light2d / hist)
        for i in range(len(hist)):
            print(bins_hist[i], i, light2d[i] / hist[i])
            npt.assert_almost_equal(light2d[i] / hist[i], 1, decimal=1)

    def test_draw_light_2d_linear(self):
        np.random.seed(41)
        lightProfile = LightProfile(
            profile_list=["HERNQUIST"],
            interpol_grid_num=1000,
            max_interpolate=10,
            min_interpolate=0.01,
        )
        kwargs_profile = [{"amp": 1.0, "Rs": 0.8}]
        r_list = lightProfile.draw_light_2d_linear(kwargs_profile, n=100000)
        bins = np.linspace(0.0, 1, 20)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        light2d = lightProfile.light_2d(
            R=(bins_hist[1:] + bins_hist[:-1]) / 2.0, kwargs_list=kwargs_profile
        )
        light2d_upper = (
            lightProfile.light_2d(R=bins_hist[1:], kwargs_list=kwargs_profile)
            * bins_hist[1:]
        )
        light2d_lower = (
            lightProfile.light_2d(R=bins_hist[:-1], kwargs_list=kwargs_profile)
            * bins_hist[:-1]
        )
        light2d *= (bins_hist[1:] + bins_hist[:-1]) / 2.0
        print((light2d_upper - light2d_lower) / (light2d_upper + light2d_lower) * 2)
        light2d /= np.sum(light2d)
        hist /= np.sum(hist)
        print(light2d / hist)
        for i in range(2, len(hist)):
            print(bins_hist[i], i, light2d[i] / hist[i])
            npt.assert_almost_equal(light2d[i] / hist[i], 1, decimal=1)

    def test_draw_light_PJaffe(self):
        np.random.seed(41)
        lightProfile = LightProfile(profile_list=["PJAFFE"])
        kwargs_profile = [{"amp": 1.0, "Rs": 0.5, "Ra": 0.2}]
        r_list = lightProfile.draw_light_2d(kwargs_profile, n=100000)
        bins = np.linspace(0, 2, 10)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        light2d = lightProfile.light_2d(
            R=(bins_hist[1:] + bins_hist[:-1]) / 2.0, kwargs_list=kwargs_profile
        )
        light2d *= (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light2d /= np.sum(light2d)
        hist /= np.sum(hist)
        print(light2d / hist)
        npt.assert_almost_equal(light2d[8] / hist[8], 1, decimal=1)

        lightProfile = LightProfile(
            profile_list=["PJAFFE"], min_interpolate=0.0001, max_interpolate=20.0
        )
        kwargs_profile = [{"amp": 1.0, "Rs": 0.04, "Ra": 0.02}]
        r_list = lightProfile.draw_light_2d(kwargs_profile, n=100000)
        bins = np.linspace(0.0, 0.1, 10)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        light2d = lightProfile.light_2d(
            R=(bins_hist[1:] + bins_hist[:-1]) / 2.0, kwargs_list=kwargs_profile
        )
        light2d *= (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light2d /= np.sum(light2d)
        hist /= np.sum(hist)
        print(light2d / hist)
        npt.assert_almost_equal(light2d[5] / hist[5], 1, decimal=1)
        assert hasattr(lightProfile, "_kwargs_light_circularized")
        lightProfile.delete_cache()
        if hasattr(lightProfile, "_kwargs_light_circularized"):
            assert False

    def test_draw_light_3d_hernquist(self):
        lightProfile = LightProfile(
            profile_list=["HERNQUIST"], min_interpolate=0.0001, max_interpolate=1000.0
        )
        kwargs_profile = [{"amp": 1.0, "Rs": 0.5}]
        r_list = lightProfile.draw_light_3d(
            kwargs_profile, n=1000000, new_compute=False
        )
        print(r_list, "r_list")
        # project it

        # test with draw light 2d profile routine
        # compare with 3d analytical solution vs histogram binned
        bins = np.linspace(0.0, 10, 20)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        bins_plot = (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light3d = lightProfile.light_3d(r=bins_plot, kwargs_list=kwargs_profile)
        light3d *= bins_plot**2
        light3d /= np.sum(light3d)
        hist /= np.sum(hist)
        # import matplotlib.pyplot as plt
        # plt.plot(bins_plot , light3d/light3d[5], label='3d reference Hernquist')
        # plt.plot(bins_plot, hist / hist[5], label='hist')
        # plt.legend()
        # plt.show()
        print(light3d / hist)
        # npt.assert_almost_equal(light3d / hist, 1, decimal=1)

        # compare with 2d analytical solution vs histogram binned
        # bins = np.linspace(0.1, 1, 10)
        R, x, y = velocity_util.project2d_random(np.array(r_list))
        hist_2d, bins_hist = np.histogram(R, bins=bins, density=True)
        hist_2d /= np.sum(hist_2d)
        bins_plot = (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light2d = lightProfile.light_2d(R=bins_plot, kwargs_list=kwargs_profile)
        light2d *= bins_plot**1
        light2d /= np.sum(light2d)

        light2d_finite = lightProfile.light_2d_finite(
            R=bins_plot, kwargs_list=kwargs_profile
        )
        light2d_finite *= bins_plot**1
        light2d_finite /= np.sum(light2d_finite)
        hist /= np.sum(hist)
        # import matplotlib.pyplot as plt
        # plt.plot(bins_plot, light2d/light2d[5], '--', label='2d reference Hernquist')
        # plt.plot(bins_plot, light2d_finite / light2d_finite[5], '-.', label='2d reference Hernquist finite')
        # plt.plot(bins_plot, hist_2d / hist_2d[5], label='hist')
        # plt.legend()
        # plt.show()
        print(light2d / hist_2d)

        # plt.plot(R, r_list, '.', label='test')
        # plt.legend()
        # plt.xlim([0, 0.2])
        # plt.ylim([0, 0.2])
        # plt.show()

        npt.assert_almost_equal(light2d / hist_2d, 1, decimal=1)

    def test_draw_light_3d_power_law(self):
        lightProfile = LightProfile(
            profile_list=["POWER_LAW"], min_interpolate=0.0001, max_interpolate=1000.0
        )
        kwargs_profile = [{"amp": 1.0, "gamma": 2, "e1": 0, "e2": 0}]
        r_list = lightProfile.draw_light_3d(
            kwargs_profile, n=1000000, new_compute=False
        )
        print(r_list, "r_list")
        # project it
        R, x, y = velocity_util.project2d_random(r_list)
        # test with draw light 2d profile routine

        # compare with 3d analytical solution vs histogram binned
        bins = np.linspace(0.1, 10, 10)
        hist, bins_hist = np.histogram(r_list, bins=bins, density=True)
        bins_plot = (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light3d = lightProfile.light_3d(r=bins_plot, kwargs_list=kwargs_profile)
        light3d *= bins_plot**2
        light3d /= np.sum(light3d)
        hist /= np.sum(hist)
        # import matplotlib.pyplot as plt
        # plt.plot(bins_plot , light3d/light3d[5], label='3d reference power-law')
        # plt.plot(bins_plot, hist / hist[5], label='hist')
        # plt.legend()
        # plt.show()
        print(light3d / hist)
        npt.assert_almost_equal(light3d / hist, 1, decimal=1)

        # compare with 2d analytical solution vs histogram binned
        # bins = np.linspace(0.1, 1, 10)
        hist, bins_hist = np.histogram(R, bins=bins, density=True)
        bins_plot = (bins_hist[1:] + bins_hist[:-1]) / 2.0
        light2d = lightProfile.light_2d_finite(R=bins_plot, kwargs_list=kwargs_profile)
        light2d *= bins_plot**1
        light2d /= np.sum(light2d)
        hist /= np.sum(hist)
        # import matplotlib.pyplot as plt
        # plt.plot(bins_plot , light2d/light2d[5], label='2d reference power-law')
        # plt.plot(bins_plot, hist / hist[5], label='hist')
        # plt.legend()
        # plt.show()
        print(light2d / hist)
        npt.assert_almost_equal(light2d / hist, 1, decimal=1)

    def test_ellipticity_in_profiles(self):
        np.random.seed(41)
        lightProfile = ["HERNQUIST_ELLIPSE", "PJAFFE_ELLIPSE"]
        import lenstronomy.Util.param_util as param_util

        phi, q = 0.14944144075912402, 0.4105628122365978
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)

        phi2, q2 = -0.37221683730659516, 0.70799587973181288
        e12, e22 = param_util.phi_q2_ellipticity(phi2, q2)
        center_x = -0.019983826426838536
        center_y = 0.90000011282957304
        kwargs_profile = [
            {
                "Rs": 0.16350224766074103,
                "e1": e1,
                "e2": e2,
                "center_x": center_x,
                "center_y": center_y,
                "amp": 1.3168943578511678,
            },
            {
                "Rs": 0.29187068596715743,
                "e1": e12,
                "e2": e22,
                "center_x": center_x,
                "center_y": center_y,
                "Ra": 0.020000382843298824,
                "amp": 85.948773973262391,
            },
        ]
        kwargs_options = {
            "lens_model_list": ["SPEP"],
            "lens_light_model_list": lightProfile,
        }
        lensAnalysis = LightProfileAnalysis(LightModel(light_model_list=lightProfile))
        r_eff = lensAnalysis.half_light_radius(
            kwargs_profile,
            center_x=center_x,
            center_y=center_y,
            grid_spacing=0.1,
            grid_num=100,
        )
        kwargs_profile[0]["e1"], kwargs_profile[0]["e2"] = 0, 0
        kwargs_profile[1]["e1"], kwargs_profile[1]["e2"] = 0, 0
        r_eff_spherical = lensAnalysis.half_light_radius(
            kwargs_profile,
            center_x=center_x,
            center_y=center_y,
            grid_spacing=0.1,
            grid_num=100,
        )
        npt.assert_almost_equal(r_eff / r_eff_spherical, 1, decimal=2)

    def test_light_3d(self):
        np.random.seed(41)
        lightProfile = LightProfile(profile_list=["HERNQUIST"])
        r = np.logspace(-2, 2, 100)
        kwargs_profile = [{"amp": 1.0, "Rs": 0.5}]
        light_3d = lightProfile.light_3d_interp(r, kwargs_profile)
        light_3d_exact = lightProfile.light_3d(r, kwargs_profile)
        for i in range(len(r)):
            npt.assert_almost_equal(light_3d[i] / light_3d_exact[i], 1, decimal=3)

    def test_light_2d_finite(self):
        interpol_grid_num = 5000
        max_interpolate = 10
        min_interpolate = 0.0001
        lightProfile = LightProfile(
            profile_list=["HERNQUIST"],
            interpol_grid_num=interpol_grid_num,
            max_interpolate=max_interpolate,
            min_interpolate=min_interpolate,
        )
        kwargs_profile = [{"amp": 1.0, "Rs": 1.0}]

        # check whether projected light integral is the same as analytic expression
        R = 1.0

        I_R = lightProfile.light_2d_finite(R, kwargs_profile)
        out = integrate.quad(
            lambda x: lightProfile.light_3d(np.sqrt(R**2 + x**2), kwargs_profile),
            min_interpolate,
            np.sqrt(max_interpolate**2 - R**2),
        )
        l_R_quad = out[0] * 2

        npt.assert_almost_equal(l_R_quad / I_R, 1, decimal=2)

        l_R = lightProfile.light_2d(R, kwargs_profile)
        npt.assert_almost_equal(l_R / I_R, 1, decimal=2)

    def test_del_cache(self):
        lightProfile = LightProfile(profile_list=["HERNQUIST"])
        lightProfile._light_cdf = 1
        lightProfile._light_cdf_log = 2
        lightProfile._f_light_3d = 3
        lightProfile.delete_cache()
        assert hasattr(lightProfile, "_light_cdf") is False
        assert hasattr(lightProfile, "_light_cdf_log") is False
        assert hasattr(lightProfile, "_f_light_3d") is False


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

from lenstronomy.ImSim.Numerics.numerics_subframe import NumericsSubFrame
from lenstronomy.ImSim.image2source_mapping import Image2SourceMapping
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
from lenstronomy.Util import util

import numpy as np

__all__ = ["ImageModel"]


class ImageModel(object):
    """This class uses functions of lens_model and source_model to make a lensed
    image."""

    def __init__(
        self,
        data_class,
        psf_class,
        lens_model_class=None,
        source_model_class=None,
        lens_light_model_class=None,
        point_source_class=None,
        extinction_class=None,
        kwargs_numerics=None,
        kwargs_pixelbased=None,
    ):
        """
        :param data_class: instance of ImageData() or PixelGrid() class
        :param psf_class: instance of PSF() class
        :param lens_model_class: instance of LensModel() class
        :param source_model_class: instance of LightModel() class describing the source parameters
        :param lens_light_model_class: instance of LightModel() class describing the lens light parameters
        :param point_source_class: instance of PointSource() class describing the point sources
        :param kwargs_numerics: keyword arguments with various numeric description (see ImageNumerics class for options)
        :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver
         (see SLITronomy documentation)
        """

        self.type = "single-band"
        self.num_bands = 1
        self.PSF = psf_class
        self.Data = data_class
        if hasattr(self.Data, "flux_scaling"):
            self._flux_scaling = self.Data.flux_scaling
        else:
            self._flux_scaling = 1
        self.PSF.set_pixel_size(self.Data.pixel_width)
        if kwargs_numerics is None:
            kwargs_numerics = {}
        self.ImageNumerics = NumericsSubFrame(
            pixel_grid=self.Data, psf=self.PSF, **kwargs_numerics
        )
        if lens_model_class is None:
            lens_model_class = LensModel(lens_model_list=[])
        self.LensModel = lens_model_class
        if point_source_class is None:
            point_source_class = PointSource(point_source_type_list=[])
        self.PointSource = point_source_class
        if self.PointSource._lensModel is None:
            self.PointSource.update_lens_model(lens_model_class=lens_model_class)
        x_center, y_center = self.Data.center
        self.PointSource.update_search_window(
            search_window=np.max(self.Data.width),
            x_center=x_center,
            y_center=y_center,
            min_distance=self.Data.pixel_width,
            only_from_unspecified=True,
        )
        self._psf_error_map = self.PSF.psf_error_map_bool

        if source_model_class is None:
            source_model_class = LightModel(light_model_list=[])
        self.SourceModel = source_model_class
        if lens_light_model_class is None:
            lens_light_model_class = LightModel(light_model_list=[])
        self.LensLightModel = lens_light_model_class
        self._kwargs_numerics = kwargs_numerics
        if extinction_class is None:
            extinction_class = DifferentialExtinction(optical_depth_model=[])
        self._extinction = extinction_class
        if kwargs_pixelbased is None:
            kwargs_pixelbased = {}
        else:
            kwargs_pixelbased = kwargs_pixelbased.copy()
        self._pixelbased_bool = self._detect_pixelbased_models()
        if self._pixelbased_bool is True:
            from slitronomy.Util.class_util import (
                create_solver_class,
            )  # requirement on SLITronomy is exclusively here

            self.SourceNumerics = self._setup_pixelbased_source_numerics(
                kwargs_numerics, kwargs_pixelbased
            )
            self.PixelSolver = create_solver_class(
                self.Data,
                self.PSF,
                self.ImageNumerics,
                self.SourceNumerics,
                self.LensModel,
                self.SourceModel,
                self.LensLightModel,
                self.PointSource,
                self._extinction,
                kwargs_pixelbased,
            )
            self.source_mapping = None  # handled with pixelated operator
        else:
            self.source_mapping = Image2SourceMapping(
                lensModel=lens_model_class, sourceModel=source_model_class
            )

        self._pb = data_class.primary_beam
        if self._pb is not None:
            self._pb_1d = util.image2array(self._pb)
        else:
            self._pb_1d = None

    def reset_point_source_cache(self, cache=True):
        """Deletes all the cache in the point source class and saves it from then on.

        :param cache: boolean, if True, saves the next occuring point source positions
            in the cache
        :return: None
        """
        self.PointSource.delete_lens_model_cache()
        self.PointSource.set_save_cache(cache)

    def update_psf(self, psf_class):
        """Update the instance of the class with a new instance of PSF() with a
        potentially different point spread function.

        :param psf_class:
        :return: no return. Class is updated.
        """
        self.PSF = psf_class
        self.PSF.set_pixel_size(self.Data.pixel_width)
        self.ImageNumerics = NumericsSubFrame(
            pixel_grid=self.Data, psf=self.PSF, **self._kwargs_numerics
        )

    def source_surface_brightness(
        self,
        kwargs_source,
        kwargs_lens=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        de_lensed=False,
        k=None,
        update_pixelbased_mapping=True,
    ):
        """Computes the source surface brightness distribution.

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness
            profile, otherwise the lensed.
        :param k: integer, if set, will only return the model of the specific index
        :return: 2d array of surface brightness pixels
        """
        return self._source_surface_brightness(
            kwargs_source,
            kwargs_lens,
            kwargs_extinction=kwargs_extinction,
            kwargs_special=kwargs_special,
            unconvolved=unconvolved,
            de_lensed=de_lensed,
            k=k,
            update_pixelbased_mapping=update_pixelbased_mapping,
        )

    def _source_surface_brightness(
        self,
        kwargs_source,
        kwargs_lens=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        de_lensed=False,
        k=None,
        update_pixelbased_mapping=True,
    ):
        """Computes the source surface brightness distribution.

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness
            profile, otherwise the lensed.
        :param k: integer, if set, will only return the model of the specific index
        :return: 2d array of surface brightness pixels
        """
        if len(self.SourceModel.profile_type_list) == 0:
            return np.zeros(self.Data.num_pixel_axes)
        if self._pixelbased_bool is True:
            return self._source_surface_brightness_pixelbased(
                kwargs_source,
                kwargs_lens=kwargs_lens,
                kwargs_extinction=kwargs_extinction,
                kwargs_special=kwargs_special,
                unconvolved=unconvolved,
                de_lensed=de_lensed,
                k=k,
                update_mapping=update_pixelbased_mapping,
            )
        else:
            return self._source_surface_brightness_analytical(
                kwargs_source,
                kwargs_lens=kwargs_lens,
                kwargs_extinction=kwargs_extinction,
                kwargs_special=kwargs_special,
                unconvolved=unconvolved,
                de_lensed=de_lensed,
                k=k,
            )

    def _source_surface_brightness_analytical(
        self,
        kwargs_source,
        kwargs_lens=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        de_lensed=False,
        k=None,
    ):
        """Computes the source surface brightness distribution.

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness
            profile, otherwise the lensed.
        :param k: integer, if set, will only return the model of the specific index
        :return: 2d array of surface brightness pixels
        """
        ra_grid, dec_grid = self.ImageNumerics.coordinates_evaluate
        if de_lensed is True:
            source_light = self.SourceModel.surface_brightness(
                ra_grid, dec_grid, kwargs_source, k=k
            )
        else:
            source_light = self.source_mapping.image_flux_joint(
                ra_grid, dec_grid, kwargs_lens, kwargs_source, k=k
            )
            source_light *= self._extinction.extinction(
                ra_grid,
                dec_grid,
                kwargs_extinction=kwargs_extinction,
                kwargs_special=kwargs_special,
            )

        # multiply with primary beam before convolution
        if self._pb is not None:
            source_light *= self._pb_1d

        source_light_final = self.ImageNumerics.re_size_convolve(
            source_light, unconvolved=unconvolved
        )
        return source_light_final * self._flux_scaling

    def _source_surface_brightness_pixelbased(
        self,
        kwargs_source,
        kwargs_lens=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        de_lensed=False,
        k=None,
        update_mapping=True,
    ):
        """Computes the source surface brightness distribution, using pixel-based solver
        for light profiles (from SLITronomy)

        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_extinction: list of keyword arguments of extinction model
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param de_lensed: if True: returns the un-lensed source surface brightness
            profile, otherwise the lensed.
        :param k: integer, if set, will only return the model of the specific index
        :param update_mapping: if False, prevent the pixelated lensing mapping to be
            updated (saves computation time if previously computed).
        :return: 2d array of surface brightness pixels
        """
        ra_grid, dec_grid = self.SourceNumerics.coordinates_evaluate
        source_light = self.SourceModel.surface_brightness(
            ra_grid, dec_grid, kwargs_source, k=k
        )
        if de_lensed is True:
            source_light = self.SourceNumerics.re_size_convolve(
                source_light, unconvolved=unconvolved
            )
        else:
            source_mapping = self.PixelSolver.lensingOperator
            source_light = source_mapping.source2image(
                source_light,
                kwargs_lens=kwargs_lens,
                kwargs_special=kwargs_special,
                update_mapping=update_mapping,
                original_source_grid=True,
            )
            source_light = self.ImageNumerics.re_size_convolve(
                source_light, unconvolved=unconvolved
            )
        # undo flux normalization performed by re_size_convolve (already handled in SLITronomy)
        source_light_final = source_light / self.Data.pixel_width**2
        return source_light_final * self._flux_scaling

    def lens_surface_brightness(self, kwargs_lens_light, unconvolved=False, k=None):
        """Computes the lens surface brightness distribution.

        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect
            seeing), otherwise convolved with PSF kernel
        :return: 2d array of surface brightness pixels
        """
        return self._lens_surface_brightness(
            kwargs_lens_light, unconvolved=unconvolved, k=k
        )

    def _lens_surface_brightness(self, kwargs_lens_light, unconvolved=False, k=None):
        """Computes the lens surface brightness distribution.

        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect
            seeing), otherwise convolved with PSF kernel
        :return: 2d array of surface brightness pixels
        """
        if self._pixelbased_bool is True:
            if unconvolved is True:
                raise ValueError(
                    "Lens light pixel-based modelling does not perform deconvolution"
                )
            return self._lens_surface_brightness_pixelbased(kwargs_lens_light, k=k)
        else:
            return self._lens_surface_brightness_analytical(
                kwargs_lens_light, unconvolved=unconvolved, k=k
            )

    def _lens_surface_brightness_analytical(
        self, kwargs_lens_light, unconvolved=False, k=None
    ):
        """Computes the lens surface brightness distribution.

        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param unconvolved: if True, returns unconvolved surface brightness (perfect
            seeing), otherwise convolved with PSF kernel
        :return: 2d array of surface brightness pixels
        """
        ra_grid, dec_grid = self.ImageNumerics.coordinates_evaluate
        lens_light = self.LensLightModel.surface_brightness(
            ra_grid, dec_grid, kwargs_lens_light, k=k
        )

        # multiply with primary beam before convolution
        if self._pb is not None:
            lens_light *= self._pb_1d

        lens_light_final = self.ImageNumerics.re_size_convolve(
            lens_light, unconvolved=unconvolved
        )
        return lens_light_final * self._flux_scaling

    def _lens_surface_brightness_pixelbased(self, kwargs_lens_light, k=None):
        """

        computes the lens surface brightness distribution , using pixel-based solver for light profiles (from SLITronomy)
        Important: SLITronomy does not solve for deconvolved lens light, hence the returned map is convolved with the PSF.

        :param kwargs_lens_light: list of keyword arguments corresponding to different lens light surface brightness profiles
        :return: 2d array of surface brightness pixels
        """
        ra_grid, dec_grid = self.ImageNumerics.coordinates_evaluate
        lens_light = self.LensLightModel.surface_brightness(
            ra_grid, dec_grid, kwargs_lens_light, k=k
        )
        lens_light_final = util.array2image(lens_light)
        return lens_light_final * self._flux_scaling

    def point_source(
        self,
        kwargs_ps,
        kwargs_lens=None,
        kwargs_special=None,
        unconvolved=False,
        k=None,
    ):
        """Computes the point source positions and paints PSF convolutions on them.

        :param kwargs_ps:
        :param kwargs_lens:
        :param kwargs_special:
        :param unconvolved:
        :param k:
        :return:
        """
        return self._point_source(
            kwargs_ps=kwargs_ps,
            kwargs_lens=kwargs_lens,
            kwargs_special=kwargs_special,
            unconvolved=unconvolved,
            k=k,
        )

    def _point_source(
        self,
        kwargs_ps,
        kwargs_lens=None,
        kwargs_special=None,
        unconvolved=False,
        k=None,
    ):
        """Computes the point source positions and paints PSF convolutions on them.

        :param kwargs_ps:
        :param kwargs_lens:
        :param kwargs_special:
        :param unconvolved:
        :param k:
        :return:
        """
        point_source_image = np.zeros((self.Data.num_pixel_axes))
        if unconvolved or self.PointSource is None:
            return point_source_image
        ra_pos, dec_pos, amp = self.PointSource.point_source_list(
            kwargs_ps, kwargs_lens=kwargs_lens, k=k
        )
        # raise warnings when primary beam is attempted to be applied to point sources.
        if len(ra_pos) != 0 and self._pb is not None:
            raise Warning(
                "Antenna primary beam does not apply to point sources in ImageModel!"
            )
        ra_pos, dec_pos = self._displace_astrometry(
            ra_pos, dec_pos, kwargs_special=kwargs_special
        )
        point_source_image += self.ImageNumerics.point_source_rendering(
            ra_pos, dec_pos, amp
        )
        return point_source_image * self._flux_scaling

    def image(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        source_add=True,
        lens_light_add=True,
        point_source_add=True,
    ):
        """Make an image with a realisation of linear parameter values "param".

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, add point sources, otherwise without
        :return: 2d array of surface brightness pixels of the simulation
        """
        return self._image(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction,
            kwargs_special,
            unconvolved,
            source_add,
            lens_light_add,
            point_source_add,
        )

    def _image(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        unconvolved=False,
        source_add=True,
        lens_light_add=True,
        point_source_add=True,
    ):
        """Make an image with a realisation of linear parameter values "param".

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param unconvolved: if True: returns the unconvolved light distribution (prefect
            seeing)
        :param source_add: if True, compute source, otherwise without
        :param lens_light_add: if True, compute lens light, otherwise without
        :param point_source_add: if True, add point sources, otherwise without
        :return: 2d array of surface brightness pixels of the simulation
        """
        model = np.zeros(self.Data.num_pixel_axes)
        if source_add is True:
            model += self._source_surface_brightness(
                kwargs_source,
                kwargs_lens,
                kwargs_extinction=kwargs_extinction,
                kwargs_special=kwargs_special,
                unconvolved=unconvolved,
            )
        if lens_light_add is True:
            model += self._lens_surface_brightness(
                kwargs_lens_light, unconvolved=unconvolved
            )
        if point_source_add is True:
            model += self._point_source(
                kwargs_ps,
                kwargs_lens,
                kwargs_special=kwargs_special,
                unconvolved=unconvolved,
            )
        return model

    def extinction_map(self, kwargs_extinction=None, kwargs_special=None):
        """Differential extinction per pixel.

        :param kwargs_extinction: list of keyword arguments corresponding to the optical
            depth models tau, such that extinction is exp(-tau)
        :param kwargs_special: keyword arguments, additional parameter to the extinction
        :return: 2d array of size of the image
        """
        return self._extinction_map(kwargs_extinction, kwargs_special)

    def _extinction_map(self, kwargs_extinction=None, kwargs_special=None):
        """Differential extinction per pixel.

        :param kwargs_extinction: list of keyword arguments corresponding to the optical
            depth models tau, such that extinction is exp(-tau)
        :param kwargs_special: keyword arguments, additional parameter to the extinction
        :return: 2d array of size of the image
        """
        ra_grid, dec_grid = self.ImageNumerics.coordinates_evaluate
        extinction = self._extinction.extinction(
            ra_grid,
            dec_grid,
            kwargs_extinction=kwargs_extinction,
            kwargs_special=kwargs_special,
        )
        print(extinction, "test extinction")
        extinction_array = np.ones_like(ra_grid) * extinction
        extinction = (
            self.ImageNumerics.re_size_convolve(extinction_array, unconvolved=True)
            / self.ImageNumerics.grid_class.pixel_width**2
        )
        return extinction

    @staticmethod
    def _displace_astrometry(x_pos, y_pos, kwargs_special=None):
        """Displaces point sources by shifts specified in kwargs_special.

        :param x_pos: list of point source positions according to point source model
            list
        :param y_pos: list of point source positions according to point source model
            list
        :param kwargs_special: keyword arguments, can contain 'delta_x_image' and
            'delta_y_image' The list is defined in order of the image positions
        :return: shifted image positions in same format as input
        """
        if kwargs_special is not None:
            if "delta_x_image" in kwargs_special:
                delta_x, delta_y = (
                    kwargs_special["delta_x_image"],
                    kwargs_special["delta_y_image"],
                )
                delta_x_new = np.zeros(len(x_pos))
                delta_x_new[0 : len(delta_x)] = delta_x[:]
                delta_y_new = np.zeros(len(y_pos))
                delta_y_new[0 : len(delta_y)] = delta_y
                x_pos = x_pos + delta_x_new
                y_pos = y_pos + delta_y_new
        return x_pos, y_pos

    def _detect_pixelbased_models(self):
        """Returns True if light profiles specific to pixel-based modelling are present
        in source model list. Otherwise returns False.

        Currently, pixel-based light profiles are: 'SLIT_STARLETS',
        'SLIT_STARLETS_GEN2'.
        """
        source_model_list = self.SourceModel.profile_type_list
        if (
            "SLIT_STARLETS" in source_model_list
            or "SLIT_STARLETS_GEN2" in source_model_list
        ):
            if len(source_model_list) > 1:
                raise ValueError(
                    "'SLIT_STARLETS' or 'SLIT_STARLETS_GEN2' must be the only source model list for pixel-based modelling"
                )
            return True
        return False

    def _setup_pixelbased_source_numerics(self, kwargs_numerics, kwargs_pixelbased):
        """Check if model requirement are compatible with support pixel-based solver,
        and creates a new numerics class specifically for source plane.

        :param kwargs_numerics: keyword argument with various numeric description (see
            ImageNumerics class for options)
        :param kwargs_pixelbased: keyword argument with various settings related to the
            pixel-based solver (see SLITronomy documentation)
        """
        # check that the required convolution type is compatible with pixel-based modelling (in current implementation)
        psf_type = self.PSF.psf_type
        supersampling_convolution = kwargs_numerics.get(
            "supersampling_convolution", False
        )
        supersampling_factor = kwargs_numerics.get("supersampling_factor", 1)
        compute_mode = kwargs_numerics.get("compute_mode", "regular")
        if psf_type not in ["PIXEL", "NONE"]:
            raise ValueError(
                "Only convolution using a pixelated kernel is supported for pixel-based modelling"
            )
        if compute_mode != "regular":
            raise ValueError(
                "Only regular coordinate grid is supported for pixel-based modelling"
            )
        if supersampling_convolution is True and supersampling_factor > 1:
            raise ValueError(
                "Only non-supersampled convolution is supported for pixel-based modelling"
            )

        # set up the source numerics with a (possibily) different supersampling resolution
        supersampling_factor_source = kwargs_pixelbased.pop(
            "supersampling_factor_source", 1
        )
        kwargs_numerics_source = kwargs_numerics.copy()
        kwargs_numerics_source["supersampling_factor"] = supersampling_factor_source
        kwargs_numerics_source["compute_mode"] = "regular"
        source_numerics_class = NumericsSubFrame(
            pixel_grid=self.Data, psf=self.PSF, **kwargs_numerics_source
        )
        return source_numerics_class



from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Util import simulation_util as sim_util
from lenstronomy.Util import param_util
from lenstronomy.Util import image_util
import pytest

import numpy as np
import numpy.testing as npt
import unittest

from lenstronomy.Analysis.multi_patch_reconstruction import (
    MultiPatchReconstruction,
    _update_frame_size,
)


class TestMultiPatchReconstruction(object):
    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel (Gaussian)
        exp_time = 100.0  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.1  # full width half max of PSF (only valid when psf_type='gaussian')
        psf_type = "GAUSSIAN"  # 'GAUSSIAN', 'PIXEL', 'NONE'

        # generate the coordinate grid and image properties
        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        kwargs_data["exposure_time"] = exp_time * np.ones_like(
            kwargs_data["image_data"]
        )
        data_class = ImageData(**kwargs_data)
        # generate the psf variables

        kwargs_psf = {"psf_type": psf_type, "pixel_size": deltaPix, "fwhm": fwhm}
        # kwargs_psf = sim_util.psf_configure_simple(psf_type=psf_type, fwhm=fwhm, kernelsize=kernel_size, deltaPix=deltaPix, kernel=kernel)
        psf_class = PSF(**kwargs_psf)

        # lensing quantities
        kwargs_shear = {
            "gamma1": 0.02,
            "gamma2": -0.04,
        }  # shear values to the source plane
        kwargs_spemd = {
            "theta_E": 1.26,
            "gamma": 2.0,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": -0.1,
            "e2": 0.05,
        }  # parameters of the deflector lens model

        # the lens model is a supperposition of an elliptical lens model with external shear
        lens_model_list = ["EPL", "SHEAR"]
        kwargs_lens_true = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)

        # choice of source type
        source_type = "SERSIC"  # 'SERSIC' or 'SHAPELETS'

        source_x = 0.0
        source_y = 0.05

        # Sersic parameters in the initial simulation
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_sersic_source = {
            "amp": 1000,
            "R_sersic": 0.05,
            "n_sersic": 1,
            "e1": e1,
            "e2": e2,
            "center_x": source_x,
            "center_y": source_y,
        }
        # kwargs_else = {'sourcePos_x': source_x, 'sourcePos_y': source_y, 'quasar_amp': 400., 'gamma1_foreground': 0.0, 'gamma2_foreground':-0.0}
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source_true = [kwargs_sersic_source]
        source_model_class = LightModel(light_model_list=source_model_list)

        lensEquationSolver = LensEquationSolver(lens_model_class)
        x_image, y_image = lensEquationSolver.findBrightImage(
            source_x,
            source_y,
            kwargs_lens_true,
            numImages=4,
            min_distance=deltaPix,
            search_window=numPix * deltaPix,
        )
        mag = lens_model_class.magnification(x_image, y_image, kwargs=kwargs_lens_true)

        kwargs_numerics = {"supersampling_factor": 1}

        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            kwargs_numerics=kwargs_numerics,
        )

        # generate image
        model = imageModel.image(kwargs_lens_true, kwargs_source_true)
        poisson = image_util.add_poisson(model, exp_time=exp_time)
        bkg = image_util.add_background(model, sigma_bkd=sigma_bkg)
        image_sim = model + bkg + poisson

        data_class.update_data(image_sim)
        kwargs_data["image_data"] = image_sim

        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
        }

        # make cutous and data instances of them
        x_pos, y_pos = data_class.map_coord2pix(x_image, y_image)
        ra_grid, dec_grid = data_class.pixel_coordinates

        multi_band_list = []
        for i in range(len(x_pos)):
            n_cut = 12
            x_c = int(x_pos[i])
            y_c = int(y_pos[i])
            image_cut = image_sim[
                int(y_c - n_cut) : int(y_c + n_cut), int(x_c - n_cut) : int(x_c + n_cut)
            ]
            exposure_map_cut = data_class.exposure_map[
                int(y_c - n_cut) : int(y_c + n_cut), int(x_c - n_cut) : int(x_c + n_cut)
            ]
            kwargs_data_i = {
                "background_rms": data_class.background_rms,
                "exposure_time": exposure_map_cut,
                "ra_at_xy_0": ra_grid[y_c - n_cut, x_c - n_cut],
                "dec_at_xy_0": dec_grid[y_c - n_cut, x_c - n_cut],
                "transform_pix2angle": data_class.transform_pix2angle,
                "image_data": image_cut,
            }
            multi_band_list.append([kwargs_data_i, kwargs_psf, kwargs_numerics])

        kwargs_params = {
            "kwargs_lens": kwargs_lens_true,
            "kwargs_source": kwargs_source_true,
        }
        self.multiPatch = MultiPatchReconstruction(
            multi_band_list,
            kwargs_model,
            kwargs_params,
            multi_band_type="joint-linear",
            kwargs_likelihood=None,
            verbose=True,
        )
        self.data_class = data_class
        self.model = model
        self.lens_model_class = lens_model_class
        self.kwargs_lens = kwargs_lens_true

        # test multi_patch with initial pixel grid
        kwargs_pixel_grid = {
            "nx": numPix,
            "ny": numPix,
            "transform_pix2angle": kwargs_data["transform_pix2angle"],
            "ra_at_xy_0": kwargs_data["ra_at_xy_0"],
            "dec_at_xy_0": kwargs_data["dec_at_xy_0"],
        }
        multiPatch = MultiPatchReconstruction(
            multi_band_list,
            kwargs_model,
            kwargs_params,
            multi_band_type="joint-linear",
            kwargs_pixel_grid=kwargs_pixel_grid,
        )

    def test_pixel_grid(self):
        pixel_grid = self.multiPatch.pixel_grid_joint
        nx, ny = pixel_grid.num_pixel_axes
        assert nx == 74
        assert ny == 67

    def test_image_joint(self):
        image_joint, model_joint, norm_residuals_joint = self.multiPatch.image_joint()

        # compute pixel shift from original
        pixel_grid = self.multiPatch.pixel_grid_joint
        nx, ny = pixel_grid.num_pixel_axes
        ra, dec = pixel_grid.radec_at_xy_0
        x0, y0 = self.data_class.map_coord2pix(ra, dec)
        # cutout original
        data = self.data_class.data
        data_cut = data[int(y0) : int(y0 + ny), int(x0) : int(x0 + nx)]
        model_cut = self.model[int(y0) : int(y0 + ny), int(x0) : int(x0 + nx)]
        # compare with original
        npt.assert_almost_equal(
            data_cut[image_joint > 0], image_joint[image_joint > 0], decimal=5
        )
        model_cut[model_joint == 0] = 0
        print(np.sum(model_cut), np.sum(model_joint), "test sum")
        # import matplotlib.pyplot as plt
        # plt.matshow((model_joint - model_cut))
        # plt.show()

        # plt.matshow(model_cut)
        # plt.show()
        # TODO make this test more precise (to do with narrower PSF convolution?)
        npt.assert_almost_equal(
            model_cut[model_joint > 0], model_joint[model_joint > 0], decimal=1
        )

    def test_lens_model_joint(self):
        (
            kappa_joint,
            magnification_joint,
            alpha_x_joint,
            alpha_y_joint,
        ) = self.multiPatch.lens_model_joint()

        # compute pixel shift from original
        pixel_grid = self.multiPatch.pixel_grid_joint
        nx, ny = pixel_grid.num_pixel_axes
        ra, dec = pixel_grid.radec_at_xy_0
        x0, y0 = self.data_class.map_coord2pix(ra, dec)
        # cutout original
        x_grid, y_grid = self.data_class.pixel_coordinates
        kappa = self.lens_model_class.kappa(x_grid, y_grid, self.kwargs_lens)
        kappa_cut = kappa[int(y0) : int(y0 + ny), int(x0) : int(x0 + nx)]
        # compare with original
        npt.assert_almost_equal(
            kappa_cut[kappa_joint > 0], kappa_joint[kappa_joint > 0], decimal=5
        )

        alpha_x, alpha_y = self.lens_model_class.alpha(x_grid, y_grid, self.kwargs_lens)
        alpha_x_cut = alpha_x[int(y0) : int(y0 + ny), int(x0) : int(x0 + nx)]
        # compare with original
        npt.assert_almost_equal(
            alpha_x_cut[alpha_x_joint > 0], alpha_x_joint[alpha_x_joint > 0], decimal=5
        )

    def test_source(self):
        source, coords = self.multiPatch.source(num_pix=50, delta_pix=0.01, center=None)
        nx, ny = np.shape(source)
        assert nx == 50

        source, coords = self.multiPatch.source(
            num_pix=50, delta_pix=0.01, center=[0, 0]
        )
        nx, ny = np.shape(source)
        assert nx == 50

    def test__update_frame_size(self):
        nx, ny = _update_frame_size(nx=10, ny=10, x_min=-5, y_min=2, nx_i=5, ny_i=5)
        assert nx == 15


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            MultiPatchReconstruction(
                multi_band_list=[],
                kwargs_model={},
                kwargs_params={},
                multi_band_type="multi-linear",
                kwargs_likelihood=None,
                verbose=True,
            )


if __name__ == "__main__":
    pytest.main()



"""Tests for `Galkin` module."""
import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.Util import constants as const
import scipy.integrate as integrate

from lenstronomy.GalKin.numeric_kinematics import NumericKinematics
from lenstronomy.GalKin.analytic_kinematics import AnalyticKinematics


class TestMassProfile(object):
    def setup_method(self):
        pass

    def test_mass_3d(self):
        kwargs_model = {
            "mass_profile_list": ["HERNQUIST"],
            "light_profile_list": ["HERNQUIST"],
            "anisotropy_model": "isotropic",
        }
        massProfile = NumericKinematics(
            kwargs_model=kwargs_model,
            kwargs_cosmo={"d_d": 1.0, "d_s": 2.0, "d_ds": 1.0},
        )
        r = 0.3
        kwargs_profile = [{"sigma0": 1.0, "Rs": 0.5}]
        mass_3d = massProfile._mass_3d_interp(r, kwargs_profile)
        mass_3d_exact = massProfile.mass_3d(r, kwargs_profile)
        npt.assert_almost_equal(mass_3d / mass_3d_exact, 1.0, decimal=3)

    def test_sigma_r2(self):
        """Tests the solution of the Jeans equation for sigma**2(r), where r is the 3d
        radius. Test is compared to analytic OM solution with power-law and Hernquist
        light profile.

        :return:
        """
        light_profile_list = ["HERNQUIST"]
        r_eff = 0.5
        Rs = 0.551 * r_eff
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 1.95
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 0.5
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 4000,
            "min_integrate": 0.001,
        }

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        analytic_kin = AnalyticKinematics(kwargs_cosmo, **kwargs_numerics)
        numeric_kin = NumericKinematics(kwargs_model, kwargs_cosmo, **kwargs_numerics)
        rho0_r0_gamma = analytic_kin._rho0_r0_gamma(theta_E, gamma)
        r_array = np.logspace(-2.9, 2.9, 100)
        sigma_r_analytic_array = []
        sigma_r_num_array = []
        for r in r_array:
            sigma_r2_analytic = analytic_kin._sigma_r2(
                r=r, a=Rs, gamma=gamma, r_ani=r_ani, rho0_r0_gamma=rho0_r0_gamma
            )
            sigma_r2_num = numeric_kin.sigma_r2(
                r, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            sigma_r_analytic = np.sqrt(sigma_r2_analytic) / 1000
            sigma_r_num = np.sqrt(sigma_r2_num) / 1000
            sigma_r_num_array.append(sigma_r_num)
            sigma_r_analytic_array.append(sigma_r_analytic)

        npt.assert_almost_equal(sigma_r_num_array, sigma_r_analytic_array, decimal=-2)
        npt.assert_almost_equal(
            np.array(sigma_r_num_array) / np.array(sigma_r_analytic_array),
            1,
            decimal=-2,
        )
        print(np.array(sigma_r_num_array) / np.array(sigma_r_analytic_array))

    def test_lum_weighted_vel_disp(self):
        """Test luminosity weighted velocity dispersion."""
        light_profile_list = ["HERNQUIST"]
        r_eff = 0.5
        Rs = 0.551 * r_eff
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 1.95
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 0.5
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 4000,
            "min_integrate": 0.001,
        }

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        # analytic_kin = AnalyticKinematics(kwargs_cosmo, **kwargs_numerics)
        numeric_kin = NumericKinematics(kwargs_model, kwargs_cosmo, **kwargs_numerics)

        R = 1
        vel_disp = numeric_kin.lum_weighted_vel_disp(
            R, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        npt.assert_almost_equal(vel_disp, 244, decimal=-1)

    def test_sigma_s2(self):
        """Test LOS projected velocity dispersion at 3d ratios (numerical Jeans equation
        solution vs analytic one)"""
        light_profile_list = ["HERNQUIST"]
        r_eff = 0.5
        Rs = 0.551 * r_eff
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 1.95
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 0.5
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 4000,
            "min_integrate": 0.001,
        }

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        analytic_kin = AnalyticKinematics(kwargs_cosmo, **kwargs_numerics)
        numeric_kin = NumericKinematics(kwargs_model, kwargs_cosmo, **kwargs_numerics)
        r_list = np.logspace(-2, 1, 10)
        for r in r_list:
            for R in np.linspace(start=0, stop=r, num=5):
                sigma_s2_analytic, I_R = analytic_kin.sigma_s2(
                    r,
                    R,
                    {"theta_E": theta_E, "gamma": gamma},
                    {"r_eff": r_eff},
                    kwargs_anisotropy,
                )
                sigma_s2_full_num = numeric_kin.sigma_s2_r(
                    r, R, kwargs_mass, kwargs_light, kwargs_anisotropy
                )
                npt.assert_almost_equal(
                    sigma_s2_full_num / sigma_s2_analytic, 1, decimal=2
                )

    def test_I_R_sigma_s2(self):
        light_profile_list = ["HERNQUIST"]
        r_eff = 1
        Rs = 0.551 * r_eff
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 1.95
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 0.5
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 4000,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.0001,
            "max_light_draw": 50,
        }

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        numeric_kin = NumericKinematics(kwargs_model, kwargs_cosmo, **kwargs_numerics)

        # check whether projected light integral is the same as analytic expression
        R = 1
        I_R_sigma2, I_R = numeric_kin._I_R_sigma2(
            R, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        out = integrate.quad(
            lambda x: numeric_kin.lightProfile.light_3d(
                np.sqrt(R**2 + x**2), kwargs_light
            ),
            kwargs_numerics["min_integrate"],
            np.sqrt(kwargs_numerics["max_integrate"] ** 2 - R**2),
        )
        l_R_quad = out[0] * 2
        npt.assert_almost_equal(l_R_quad / I_R, 1, decimal=2)

        l_R = numeric_kin.lightProfile.light_2d(R, kwargs_light)
        npt.assert_almost_equal(l_R / I_R, 1, decimal=2)

    def test_log_linear_integral(self):
        # light profile
        light_profile_list = ["HERNQUIST"]
        Rs = 0.5
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics_linear = {
            "interpol_grid_num": 2000,
            "log_integration": False,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }
        kwargs_numerics_log = {
            "interpol_grid_num": 1000,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }

        numerics_linear = NumericKinematics(
            kwargs_model=kwargs_model,
            kwargs_cosmo=kwargs_cosmo,
            **kwargs_numerics_linear
        )
        numerics_log = NumericKinematics(
            kwargs_model=kwargs_model, kwargs_cosmo=kwargs_cosmo, **kwargs_numerics_log
        )
        R = np.logspace(-2, 0, 100)

        lin_I_R = np.zeros_like(R)
        log_I_R = np.zeros_like(R)
        for i in range(len(R)):
            lin_I_R[i], _ = numerics_linear._I_R_sigma2(
                R[i], kwargs_profile, kwargs_light, kwargs_anisotropy
            )
            log_I_R[i], _ = numerics_log._I_R_sigma2(
                R[i], kwargs_profile, kwargs_light, kwargs_anisotropy
            )

        # import matplotlib.pyplot as plt
        # plt.semilogx(R, lin_I_R / log_I_R, 'r', label='lin /log integrate')
        # plt.legend()
        # plt.show()

        R_ = 1
        r_array = np.logspace(start=np.log10(R_ + 0.001), stop=1, num=100)
        integrad_a15 = numerics_linear._integrand_A15(
            r_array, R_, kwargs_profile, kwargs_light, kwargs_anisotropy
        )
        # plt.loglog(r_array, integrad_a15)
        # plt.show()

        for i in range(len(R)):
            npt.assert_almost_equal(log_I_R[i] / lin_I_R[i], 1, decimal=2)

        # assert 1 == 0

    def test_I_R_sigma(self):
        """Test numerical integral against quad integrator :return:"""
        light_profile_list = ["HERNQUIST"]
        Rs = 0.5
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }

        numerics = NumericKinematics(
            kwargs_model=kwargs_model, kwargs_cosmo=kwargs_cosmo, **kwargs_numerics
        )

        R = 0.1
        out = integrate.quad(
            lambda x: numerics._integrand_A15(
                x, R, kwargs_profile, kwargs_light, kwargs_anisotropy
            ),
            R,
            kwargs_numerics["max_integrate"],
        )

        I_R_sigma_quad = (
            out[0] * 2 * const.G / (const.arcsec * kwargs_cosmo["d_d"] * const.Mpc)
        )
        I_R_sigma_numerics_log, _ = numerics._I_R_sigma2(
            R, kwargs_profile, kwargs_light, kwargs_anisotropy
        )

        kwargs_numerics_lin = {
            "interpol_grid_num": 2000,
            "log_integration": False,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }
        numerics_lin = NumericKinematics(
            kwargs_model=kwargs_model, kwargs_cosmo=kwargs_cosmo, **kwargs_numerics_lin
        )
        I_R_simga_numerics_lin, _ = numerics_lin._I_R_sigma2(
            R, kwargs_profile, kwargs_light, kwargs_anisotropy
        )
        npt.assert_almost_equal(I_R_sigma_numerics_log / I_R_sigma_quad, 1, decimal=2)

        # We do not test the linear integral as it is not as accurate!!!
        # npt.assert_almost_equal(I_R_simga_numerics_lin / I_R_simga_quad, 1, decimal=2)

        # here we test the interpolation
        numerics = NumericKinematics(
            kwargs_model=kwargs_model, kwargs_cosmo=kwargs_cosmo, **kwargs_numerics
        )
        R = 1
        I_R_sigma2_interp, I_R_interp = numerics._I_R_sigma2_interp(
            R, kwargs_profile, kwargs_light, kwargs_anisotropy
        )
        I_R_sigma2, I_R = numerics._I_R_sigma2(
            R, kwargs_profile, kwargs_light, kwargs_anisotropy
        )
        npt.assert_almost_equal(I_R_sigma2_interp / I_R_sigma2, 1, decimal=3)
        npt.assert_almost_equal(I_R_interp / I_R, 1, decimal=3)

    def test_power_law_test(self):
        # tests a isotropic velocity anisotropy on a singular isothermal sphere with the same tracer particle distribution
        # This should result in a constant velocity dispersion as a function of radius, analytically known

        # set up power-law light profile
        light_model = ["POWER_LAW"]
        kwargs_light = [{"gamma": 2, "amp": 1, "e1": 0, "e2": 0}]

        lens_model = ["SIS"]
        kwargs_mass = [{"theta_E": 1}]

        anisotropy_type = "isotropic"
        kwargs_anisotropy = {}
        kwargs_model = {
            "mass_profile_list": lens_model,
            "light_profile_list": light_model,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}

        # compute analytic velocity dispersion of SIS profile

        v_sigma_c2 = (
            kwargs_mass[0]["theta_E"]
            * const.arcsec
            / (4 * np.pi)
            * kwargs_cosmo["d_s"]
            / kwargs_cosmo["d_ds"]
        )
        v_sigma_true = np.sqrt(v_sigma_c2) * const.c / 1000

        numerics = NumericKinematics(
            kwargs_model=kwargs_model, kwargs_cosmo=kwargs_cosmo, **kwargs_numerics
        )
        R = 2
        I_R_sigma2, I_R = numerics._I_R_sigma2(
            R, kwargs_mass, kwargs_light, kwargs_anisotropy={}
        )
        sigma_v = np.sqrt(I_R_sigma2 / I_R) / 1000
        print(sigma_v, v_sigma_true)
        npt.assert_almost_equal(sigma_v / v_sigma_true, 1, decimal=2)

        # plot as radial distance of projected dispersion
        r_array = np.logspace(start=-2, stop=1, num=100)

        sigma_array = np.zeros_like(r_array)
        for i, R in enumerate(r_array):
            I_R_sigma2, I_R = numerics._I_R_sigma2(
                R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            sigma_array[i] = np.sqrt(I_R_sigma2 / I_R) / 1000

        # import matplotlib.pyplot as plt
        # plt.semilogx(r_array, sigma_array)
        # plt.hlines(v_sigma_true, xmin=r_array[0], xmax=r_array[-1])
        # plt.show()

        npt.assert_almost_equal(sigma_array / v_sigma_true, 1, decimal=2)

        # and here we test the 3d radial velocity dispersion solution
        sigma_array = np.zeros_like(r_array)
        R_test = 0
        for i, r in enumerate(r_array):
            sigma_s2 = numerics.sigma_s2_r(
                r, R_test, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            sigma_array[i] = np.sqrt(sigma_s2) / 1000
        npt.assert_almost_equal(sigma_array / v_sigma_true, 1, decimal=2)

    def test_delete_cache(self):
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }
        kwargs_model = {
            "mass_profile_list": [],
            "light_profile_list": [],
            "anisotropy_model": "const",
        }
        numeric_kin = NumericKinematics(kwargs_model, kwargs_cosmo, **kwargs_numerics)
        numeric_kin._interp_jeans_integral = 1
        numeric_kin._log_mass_3d = 2
        numeric_kin.delete_cache()
        assert hasattr(numeric_kin, "_log_mass_3d") is False
        assert hasattr(numeric_kin, "_interp_jeans_integral") is False


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

from lenstronomy.GalKin.aperture_types import Shell, Slit, IFUShells, Frame

__all__ = ["Aperture"]
"""Class that defines the aperture of the measurement (e.g. slit, integral field
spectroscopy regions etc).

Available aperture types:
-------------------------

'slit': length, width, center_ra, center_dec, angle
'shell': r_in, r_out, center_ra, center_dec
"""


class Aperture(object):
    """Defines mask(s) of spectra, can handle IFU and single slit/box type data."""

    def __init__(self, aperture_type, **kwargs_aperture):
        """

        :param aperture_type: string
        :param kwargs_aperture: keyword arguments reflecting the aperture type chosen.
         We refer to the specific class instances for documentation.
        """
        if aperture_type == "slit":
            self._aperture = Slit(**kwargs_aperture)
        elif aperture_type == "shell":
            self._aperture = Shell(**kwargs_aperture)
        elif aperture_type == "IFU_shells":
            self._aperture = IFUShells(**kwargs_aperture)
        elif aperture_type == "frame":
            self._aperture = Frame(**kwargs_aperture)
        else:
            raise ValueError(
                "aperture type %s not implemented! Available are 'slit', 'shell', 'IFU_shells'. "
                % aperture_type
            )
        self.aperture_type = aperture_type

    def aperture_select(self, ra, dec):
        """

        :param ra: angular coordinate of photon/ray
        :param dec: angular coordinate of photon/ray
        :return: bool, True if photon/ray is within the slit, False otherwise, int of the segment of the IFU
        """
        return self._aperture.aperture_select(ra, dec)

    @property
    def num_segments(self):
        return self._aperture.num_segments



__author__ = "sibirrer"

from astropy.cosmology import FlatLambdaCDM, LambdaCDM
from lenstronomy.Cosmo.lens_cosmo import LensCosmo

__all__ = ["LCDM"]


class LCDM(object):
    """Flat LCDM cosmology background with free Hubble parameter and Omega_m at fixed
    lens redshift configuration."""

    def __init__(self, z_lens, z_source, flat=True):
        """

        :param z_lens: redshift of lens
        :param z_source: redshift of source
        :param flat: bool, if True, flat universe is assumed
        """
        self.z_lens = z_lens
        self.z_source = z_source
        self._flat = flat

    def _get_cosom(self, H_0, Om0, Ode0=None):
        """

        :param H_0:
        :param Om0:
        :param Ode0:
        :return:
        """
        if self._flat is True:
            cosmo = FlatLambdaCDM(H0=H_0, Om0=Om0)
        else:
            cosmo = LambdaCDM(H0=H_0, Om0=Om0, Ode0=Ode0)
        lensCosmo = LensCosmo(z_lens=self.z_lens, z_source=self.z_source, cosmo=cosmo)
        return lensCosmo

    def D_d(self, H_0, Om0, Ode0=None):
        """Angular diameter to deflector.

        :param H_0: Hubble parameter [km/s/Mpc]
        :param Om0: normalized matter density at present time
        :return: float [Mpc]
        """
        lensCosmo = self._get_cosom(H_0, Om0, Ode0)
        return lensCosmo.dd

    def D_s(self, H_0, Om0, Ode0=None):
        """Angular diameter to source.

        :param H_0: Hubble parameter [km/s/Mpc]
        :param Om0: normalized matter density at present time
        :return: float [Mpc]
        """
        lensCosmo = self._get_cosom(H_0, Om0, Ode0)
        return lensCosmo.ds

    def D_ds(self, H_0, Om0, Ode0=None):
        """Angular diameter from deflector to source.

        :param H_0: Hubble parameter [km/s/Mpc]
        :param Om0: normalized matter density at present time
        :return: float [Mpc]
        """
        lensCosmo = self._get_cosom(H_0, Om0, Ode0)
        return lensCosmo.dds

    def D_dt(self, H_0, Om0, Ode0=None):
        """Time-delay distance.

        :param H_0: Hubble parameter [km/s/Mpc]
        :param Om0: normalized matter density at present time
        :return: float [Mpc]
        """
        lensCosmo = self._get_cosom(H_0, Om0, Ode0)
        return lensCosmo.ddt



__author__ = "dgilman"

import unittest
from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC
from lenstronomy.LensModel.Profiles.general_nfw import GNFW
import numpy as np
import numpy.testing as npt
import pytest


class TestTNFWC(object):
    def setup_method(self):
        self.tnfwc = TNFWC()
        self.gnfw = GNFW()

    def test_alphaRs(self):
        kwargs_lens = {
            "alpha_Rs": 2.1,
            "Rs": 1.5,
            "r_core": 1.2,
            "r_trunc": 3.0,
            "center_x": 0.04,
            "center_y": -1.0,
        }
        alpha_rs = self.tnfwc.derivatives(
            kwargs_lens["Rs"],
            0.0,
            kwargs_lens["Rs"],
            kwargs_lens["alpha_Rs"],
            kwargs_lens["r_core"],
            kwargs_lens["r_trunc"],
        )[0]
        npt.assert_almost_equal(alpha_rs, kwargs_lens["alpha_Rs"], 8)

    def test_alphaRs_rho0_conversion(self):
        kwargs_lens = {
            "alpha_Rs": 2.1,
            "Rs": 1.5,
            "r_core": 1.2,
            "r_trunc": 3.0,
            "center_x": 0.04,
            "center_y": -1.0,
        }
        rho0 = self.tnfwc.alpha2rho0(
            kwargs_lens["alpha_Rs"],
            kwargs_lens["Rs"],
            kwargs_lens["r_core"],
            kwargs_lens["r_trunc"],
        )
        alpha_Rs = self.tnfwc.rho02alpha(
            rho0, kwargs_lens["Rs"], kwargs_lens["r_core"], kwargs_lens["r_trunc"]
        )
        npt.assert_almost_equal(alpha_Rs, kwargs_lens["alpha_Rs"], 5)

    def test_gnfw_match(self):
        # profile reduces to GNFW with gamma_inner = 1.0, gamma_outer = 3.0 when core -> 0 and truncation -> inf
        alpha_Rs = 2.5
        Rs = 1.5
        R = np.logspace(-1, 1, 100) * Rs
        r_core = 0.001 * Rs
        r_trunc = 1000 * Rs

        alpha_tnfwc = self.tnfwc.derivatives(R, 0.0, Rs, alpha_Rs, r_core, r_trunc)
        alpha_gnfw = self.gnfw.derivatives(R, 0.0, Rs, alpha_Rs, 1.0, 3.0)
        npt.assert_almost_equal(alpha_gnfw, alpha_tnfwc, 3.0)

        rho0 = self.tnfwc.alpha2rho0(alpha_Rs, Rs, r_core, r_trunc)
        density_2d_tnfwc = self.tnfwc.density_2d(R, 0.0, Rs, rho0, r_core, r_trunc)
        rho0 = self.gnfw.alpha2rho0(alpha_Rs, Rs, 1.0, 3.0)
        density_2d_gnfw = self.gnfw.density_2d(R, 0.0, Rs, rho0, 1.0, 3.0)
        npt.assert_almost_equal(density_2d_tnfwc, density_2d_gnfw, 3.0)

    def test_mass3d(self):
        kwargs_lens = {"alpha_Rs": 2.1, "Rs": 1.5, "r_core": 1.2, "r_trunc": 3.0}
        rho0 = self.tnfwc.alpha2rho0(
            kwargs_lens["alpha_Rs"],
            kwargs_lens["Rs"],
            kwargs_lens["r_core"],
            kwargs_lens["r_trunc"],
        )
        kwargs_lens_rho0 = {"rho0": rho0, "Rs": 1.5, "r_core": 1.2, "r_trunc": 3.0}
        m1 = self.tnfwc.mass_3d(2.0, **kwargs_lens_rho0)
        m2 = self.tnfwc.mass_3d_lens(2.0, **kwargs_lens)
        npt.assert_almost_equal(m1, m2)

    def test_g(self):
        x = 1.2
        b = 1.5
        t = 1.5
        out1 = self.tnfwc._g(x, b, t)
        b = 1.5001
        out2 = self.tnfwc._g(x, b, t)
        out3 = self.tnfwc._g(x, b, t + 0.001)
        npt.assert_almost_equal(out1, out2, 4)
        npt.assert_almost_equal(out1, out3, 4)


if __name__ == "__main__":
    pytest.main()



import pytest
import unittest

from lenstronomy.PointSource.Types.base_ps import PSBase
from lenstronomy.LensModel.lens_model import LensModel


class TestPSBase(object):
    def setup_method(self):
        self.base = PSBase(
            lens_model=LensModel(lens_model_list=[]),
            fixed_magnification=False,
            additional_images=False,
        )
        PSBase(fixed_magnification=True, additional_images=True)

    def test_update_lens_model(self):
        self.base.update_lens_model(lens_model_class=None)
        assert self.base._solver is None

        base = PSBase()
        base.update_lens_model(lens_model_class=LensModel(lens_model_list=["SIS"]))
        assert base._solver is not None
        PSBase(fixed_magnification=True, additional_images=True)


class TestUtil(object):
    def setup_method(self):
        pass

    def test_expand_to_array(self):
        from lenstronomy.PointSource.Types.base_ps import _expand_to_array

        array = 1
        num = 3
        array_out = _expand_to_array(array, num)
        assert len(array_out) == num

        array = [1]
        num = 3
        array_out = _expand_to_array(array, num)
        assert len(array_out) == num
        assert array_out[1] == 0

        array = [1, 1, 1]
        num = 3
        array_out = _expand_to_array(array, num)
        assert len(array_out) == num
        assert array_out[1] == 1

    def test_shrink_array(self):
        from lenstronomy.PointSource.Types.base_ps import _shrink_array

        array = [1, 2, 3]
        num = 2
        array_out = _shrink_array(array, num)
        assert len(array_out) == num
        assert array_out[1] == 2

        array = 1
        num = 3
        array_out = _shrink_array(array, num)
        assert array_out == array

        array = [1]
        num = 2
        with pytest.raises(ValueError):
            _shrink_array(array, num)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        base = PSBase()
        with self.assertRaises(ValueError):
            base.image_position(kwargs_ps=None)
        with self.assertRaises(ValueError):
            base.source_position(kwargs_ps=None)
        with self.assertRaises(ValueError):
            base.image_amplitude(kwargs_ps=None)
        with self.assertRaises(ValueError):
            base.source_amplitude(kwargs_ps=None)


if __name__ == "__main__":
    pytest.main()



__author__ = "dgilman", "sibirrer"

from lenstronomy.LensModel.Profiles.cnfw import CNFW
from lenstronomy.LensModel.Profiles.nfw import NFW

import numpy as np
import numpy.testing as npt
import pytest


class Testcnfw(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.cn = CNFW()
        self.n = NFW()

    def test_pot(self):
        # this test requires that the CNFW profile with a very small core results in the potential of the NFW profile
        pot1 = self.cn.function(x=2, y=0, Rs=1, alpha_Rs=1, r_core=0.001)
        pot2 = self.n.function(x=2, y=0, Rs=1, alpha_Rs=1)
        npt.assert_almost_equal(pot1 / pot2, 1, decimal=3)

    def _kappa_integrand(self, x, y, Rs, m0, r_core):
        return 2 * np.pi * x * self.cn.density_2d(x, y, Rs, m0, r_core)

    def test_derivatives(self):
        Rs = 10.0
        rho0 = 1.0
        r_core = 7.0

        R = np.linspace(0.1 * Rs, 4 * Rs, 1000)

        alpha_Rs = self.cn._rho2alpha(rho0, Rs, r_core)
        alpha = self.cn.alpha_r(R, Rs, rho0, r_core)
        alpha_theory = self.cn.mass_2d(R, Rs, rho0, r_core) / np.pi / R
        alpha_derivatives = self.cn.derivatives(R, 0, Rs, alpha_Rs, r_core)[0]

        npt.assert_almost_equal(alpha_derivatives / alpha_theory, 1)
        npt.assert_almost_equal(alpha / alpha_theory, 1)
        npt.assert_almost_equal(alpha / alpha_derivatives, 1)

    def test_mass_3d(self):
        Rs = 10.0
        rho0 = 1.0
        r_core = 7.0

        R = np.linspace(0.1 * Rs, 4 * Rs, 1000)
        alpha_Rs = self.cn._rho2alpha(rho0, Rs, r_core)
        m3d = self.cn.mass_3d(R, Rs, rho0, r_core)
        m3d_lens = self.cn.mass_3d_lens(R, Rs, alpha_Rs, r_core)
        npt.assert_almost_equal(m3d, m3d_lens, decimal=8)

    def test_mproj(self):
        Rs = 10.0
        r_core = 0.7 * Rs
        Rmax = np.linspace(0.6 * Rs, 1.1 * Rs, 1000)
        dr = Rmax[1] - Rmax[0]
        m0 = 1

        m2d = self.cn.mass_2d(Rmax, Rs, m0, r_core)
        integrand = np.gradient(m2d, dr)
        kappa_integrand = self._kappa_integrand(Rmax, 0, Rs, m0, r_core)

        mean_diff = np.absolute(kappa_integrand - integrand) * len(Rmax) ** -1

        npt.assert_almost_equal(mean_diff, 0, decimal=3)

    def test_GF(self):
        x_array = np.array([0.5, 0.8, 1.2])
        b = 0.7
        Garray = self.cn._G(x_array, b)
        Farray = self.cn._F(x_array, b)
        for i in range(0, len(x_array)):
            npt.assert_almost_equal(Farray[i], self.cn._F(x_array[i], b))
            npt.assert_almost_equal(Garray[i], self.cn._G(x_array[i], b))

    def test_gamma(self):
        Rs = 10.0
        rho0 = 1.0
        r_core = 0.7 * Rs

        R = np.array([0.5 * Rs, 0.8 * Rs, 1.1 * Rs])

        g1_array, g2_array = self.cn.cnfwGamma(R, Rs, rho0, r_core, R, 0.6 * Rs)
        for i in range(0, len(R)):
            g1, g2 = self.cn.cnfwGamma(R[i], Rs, rho0, r_core, R[i], 0.6 * Rs)
            npt.assert_almost_equal(g1_array[i], g1)
            npt.assert_almost_equal(g2_array[i], g2)


if __name__ == "__main__":
    pytest.main()
